# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""lio.py: Module to handle commands on LIO Array."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import ast

import libsan.host.ssh
import libsan.misc.array
import libsan.misc.size
from libsan import _print


class lio:  # noqaN801
    """Class to manage LIO array"""

    host = None
    user = None
    passwd = None

    san_conf_path = None
    sa_conf_dict = None

    lio_dict: dict = {}
    map_info: list = []

    def __init__(self, hostname, username, password, timeout=None, san_dev=None):
        # san_dev is here because we need it for DELL Equalogic array class
        self.san_dev_name = san_dev
        self.host = hostname
        self.user = username
        self.passwd = password
        self.timeout = timeout
        self.san_conf_path = None
        # _print("DEBUG: LIO version is %s" % self.get_version())

    def set_san_conf_path(self, san_conf_path):
        self.san_conf_path = san_conf_path
        return True

    def set_sa_conf(self, sa_conf_dict):
        self.sa_conf_dict = sa_conf_dict
        return True

    def get_sa_conf(self):
        return self.sa_conf_dict

    def _run(self, cmd, return_output=False, verbose=True):
        _print("INFO: Connecting to %s for command:" % self.host)
        print(cmd)
        session = libsan.host.ssh.connect(self.host, user=self.user, passwd=self.passwd)
        if not session:
            _print("FAIL: Could not connect to array (%s)" % self.host)
            if return_output:
                return 127, None
            return 127

        # If return_output is True, ret will be a tuple
        ret = libsan.host.ssh.run_cmd(
            session, cmd, return_output=return_output, verbose=verbose, invoke_shell=False, timeout=self.timeout
        )
        libsan.host.ssh.disconnect(session)
        return ret

    def get_version(self):
        """Return LIO version running on array"""

        # Return the version of the first group we found
        ret, version = self._run("stlio version", return_output=True, verbose=False)
        if ret != 0:
            _print("FAIL: Getting LIO version")
            return None
        return version

    def query_all(self, recheck=False):
        if not recheck and self.lio_dict:
            return self.lio_dict

        cmd = 'python -c "import libsan.host.lio as lio;'
        cmd += 'lio.lio_query(show_output=True)"'
        # cmd = "stlio get_all_luns_info"
        ret, lio_dict = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            _print("FAIL: query_all() - could not get all info")
            print(lio_dict)
            return None

        # Converting str to dict type
        try:
            lio_dict = ast.literal_eval(lio_dict)
        except Exception as e:
            print(e)
            _print("FAIL: Could not convert lio_query output to dict")
            print(lio_dict)
            return None

        self.lio_dict = lio_dict
        return lio_dict

    @staticmethod
    def capability():
        """Indicates supported operation on array"""
        cap_dict = {}
        cap_dict["lun_info"] = True
        cap_dict["lun_query"] = True
        cap_dict["lun_create"] = True
        cap_dict["lun_map"] = True
        cap_dict["lun_remove"] = True
        cap_dict["lun_grow"] = False
        cap_dict["lun_shrink"] = False
        cap_dict["lun_trepass"] = False
        cap_dict["lun_thinp"] = False
        cap_dict["sa_ctrler_reboot"] = False
        return cap_dict

    def _lun_exist(self, lun_name):
        """Test if LUN exist
        If LUN does not exist return error
        """
        if not lun_name:
            _print("FAIL: _lun_exist - requires lun_name")
            return False

        luns_dict = self.query_all_lun_info()
        if not luns_dict:
            _print("FAIL: _lun_exist() - Could not query all lun info")
            return False

        if lun_name in list(luns_dict.keys()) and luns_dict[lun_name]["backstore_type"] == "fileio":
            return True
        # cmd = "stlio lun_info --bs_type fileio --lun_name %s" % lun_name
        # ret, output = self._run(cmd, return_output = True, verbose = False)
        # if ret != 0:
        #    return False
        return False

    @staticmethod
    def _lun_wwn2wwid(lun_info_dict):
        """From the LUN WWN is possible to get WWID"""
        if not lun_info_dict:
            return None

        if "wwn" not in lun_info_dict:
            return None
        wwid = lun_info_dict["wwn"]
        wwid = wwid.replace("-", "")
        # Just the first 26 bytes are the wwid
        wwid = wwid[:25]
        wwid = "36001405" + wwid
        return wwid

    def lun_info(self, lun_name):
        """Query detailed information of specific LUN"""
        if not lun_name:
            _print("FAIL: lun_info - requires lun_name")
            return None

        # used to match regex for each session information that we support
        # supported_lun_info = {"name"    : "^.*/(\S+) has size:",
        #                       "size_human": "^.* has size: (\S+)",
        #                       "wwn"    : "^.* has wwn: (\S+)"}

        # cmd = "stlio lun_info --bs_type fileio --lun_name %s" % lun_name
        # ret, output = self._run(cmd, return_output = True, verbose = True)
        # if ret != 0:
        #   _print("FAIL: lun_info() - lun does not exist")
        #   return None

        # if not output:
        #   return None

        # lines = output.split("\n")
        # lun_info_dict = None

        # line_id = 0
        # n_lines = len(lines)
        # while line_id < n_lines:
        #   line = lines[line_id].rstrip('\n')
        #   line_id += 1
        #   print "DEBUG parse: try %s - %s" % (line)
        #   if not line:
        #       continue

        #   for key in supported_lun_info.keys():
        #       m = re.match(supported_lun_info[key], line)
        #       if not m:
        #           continue
        #       if not lun_info_dict:
        #           lun_info_dict = {}
        #       print "Found %s: %s" % (key, m.group(1))
        #       lun_info_dict[key] = m.group(1)

        #   if "wwn" in lun_info_dict.keys():
        #       wwid = self._lun_wwn2wwid(lun_info_dict)
        #       if wwid:
        #           lun_info_dict["wwid"] = wwid
        luns_dict = self.query_all_lun_info()
        if not luns_dict:
            _print("FAIL: lun_info() - Could not query all lun info")
            return None
        if lun_name in list(luns_dict.keys()) and luns_dict[lun_name]["backstore_type"] == "fileio":
            return luns_dict[lun_name]

        _print("INFO lun_info(): Could not find LUN %s with type fileio" % lun_name)
        return None

    def query_all_luns(self):
        luns_info = self.query_all_lun_info()
        if not luns_info:
            return None

        return sorted(luns_info.keys())

    def query_all_lun_info(self, recheck=False):
        """Query all LUNs from LIO, the way we are storing the info will be a problem
        If we have LUN with same name on different backstores
        """
        lio_dict = self.query_all(recheck)
        if not lio_dict:
            _print("FAIL: query_all_lun_info() - could not get all lun info")
            print(lio_dict)
            return None
        all_lun_info_dict = {}

        if "backstores" not in list(lio_dict.keys()):
            return None

        for bs_type in list(lio_dict["backstores"].keys()):
            bs_luns = lio_dict["backstores"][bs_type]
            if not bs_luns:
                # This backstore has no LUN configured
                continue
            for lun in list(bs_luns.keys()):
                lun_info = {}
                lun_info["lun_name"] = lun
                lun_info["backstore_type"] = bs_type
                lun_info["wwid"] = None

                if "wwid" in list(bs_luns[lun].keys()):
                    lun_info["wwid"] = bs_luns[lun]["wwid"]

                if "lun_size" in list(bs_luns[lun].keys()):
                    lun_info["size"] = libsan.misc.size.size_human_2_size_bytes(bs_luns[lun]["lun_size"])
                    lun_info["size_human"] = bs_luns[lun]["lun_size"]

                if "file_path" in list(bs_luns[lun].keys()):
                    lun_info["file_path"] = bs_luns[lun]["file_path"]

                if "mapping" in list(bs_luns[lun].keys()):
                    lun_info["map_infos"] = bs_luns[lun]["mapping"]

                all_lun_info_dict[lun] = lun_info

        # infos = output.split("\n")
        # lun_regex = re.compile("(\S+)/(\S+)")
        # map_regex = re.compile("(\S+) is mapped to (\S+)\(lun(\d+)\)/(\S+)\(lun(\d+)\)")
        # details_regex = re.compile("(\S+) has size: (\S+)")
        # for info in infos:
        #   if not info:
        #       #We might get empty line
        #       continue
        #   m = lun_regex.match(info)
        #   if not m:
        #       _print("FAIL: %s does not seem to contain a lun format" % info)
        #       continue
        #   bs_type = m.group(1)
        #   lun_name = m.group(2)
        #   if lun_name not in all_lun_info_dict.keys():
        #       all_lun_info_dict[lun_name] = {}
        #   lun_info = all_lun_info_dict[lun_name]
        #   lun_info["lun_name"] = lun_name
        #   lun_info["backstore_type"] = bs_type
        #   lun_info["wwid"] = None
        #   if "map_infos" not in lun_info.keys():
        #       lun_info["map_infos"] = []

        #   m = map_regex.match(info)
        #   if m:
        #       map_infos = {}
        #       map_infos["t_wwpns"] = [m.group(2)]
        #       map_infos["t_wwpn"] = m.group(2)
        #       map_infos["h_wwpns"] = [m.group(4)]
        #       map_infos["h_wwpn"] = m.group(4)
        #       map_infos["lun_id"] = m.group(5)
        #       map_infos["t_lun_id"] = m.group(3)
        #       map_infos["lun_name"] = lun_name
        #       map_infos["backstore_type"] = bs_type
        #       lun_info["map_infos"].append(map_infos)

        #   m = details_regex.match(info)
        #   if m:
        #       lun_info["size"] = size_human_2_size_bytes(m.group(2))
        #       lun_info["size_human"] = m.group(2)
        return all_lun_info_dict

    def lun_create(self, name, size):
        """Create new LUN on array if LUN does not exist"""
        _print(f"INFO: Creating fileio LUN {name} with size {size}")
        if not name or not size:
            _print("FAIL: usage lun_create(name, size)")
            return None

        if self._lun_exist(name):
            _print("INFO: LUN %s already exist" % name)
            return None

        size_bytes = libsan.misc.size.size_human_2_size_bytes(size)
        if not size_bytes:
            _print("FAIL: lun_create() - Could not convert %s to bytes" % size)
            return None

        # cmd = "stlio lun_create --bs_type fileio --lun_name %s --lun_size %s " % (name, size_bytes)
        cmd = (
            'python -c "import libsan.host.lio as lio; '
            'lio.lio_create_backstore(\\"{}\\",\\"{}\\", \\"{}\\")"'.format(
                "fileio",
                name,
                size_bytes,
            )
        )
        ret, output = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            _print('FAIL: "%s" command' % cmd)
            print(output)
            return None
        # Need to update our volumes
        vol_dict = self.query_all_lun_info(recheck=True)
        if name not in list(vol_dict.keys()):
            self.lun_remove(name)
            _print("FAIL: thought LUN %s was created, but it was not" % name)
            print(vol_dict)
            return None

        _print("INFO: LUN %s created successfully" % name)
        return name

    def lun_remove(self, name):
        """Delete a LUN on array if LUN exists"""
        if not name:
            _print("FAIL: usage lun_remove(name)")
            return False

        if not self._lun_exist(name):
            _print("INFO: LUN %s do NOT exist" % name)
            return True

        _print("INFO: Deleting LUN %s" % name)
        cmd = 'python -c "import libsan.host.lio as lio;'
        cmd += ' lio.lio_delete_backstore(\\"{}\\",\\"{}\\")"'.format("fileio", name)
        # cmd = "stlio lun_delete --bs_type fileio --lun_name %s" % (name)
        ret, output = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            _print('FAIL: "%s" command' % cmd)
            print(output)
            return False
        # Need to update our volumes
        vol_dict = self.query_all_lun_info(recheck=True)
        if name in list(vol_dict.keys()):
            _print("FAIL: thought LUN %s was deleted, but it was not" % name)
            return False

        # When deleting backstores, mapping information is there, even if there
        # is no lun mapped to it. Should clean up this
        cmd = 'python -c "import libsan.host.lio as lio;'
        cmd += ' lio.lio_clean_up_targets()"'
        # cmd = "stlio lun_delete --bs_type fileio --lun_name %s" % (name)
        ret, output = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            _print('FAIL: "%s" command' % cmd)
            print(output)
            return False
        return True

    def lun_map(self, name):
        """Map a LUN to an h_wwpn"""
        if not name:
            _print("FAIL: usage lun_map(name)")
            return False

        if not self._lun_exist(name):
            _print("INFO: LUN %s do NOT exist" % name)
            return False

        if not self.map_info:
            _print("FAIL: lun_map() - There is no mapping info to use")
            return False

        for map_info in self.map_info:
            if "t_wwpn" not in list(map_info.keys()) or "h_wwpn" not in list(map_info.keys()):
                _print("FAIL: lun_map() - Do not know how to map using")
                print(map_info)
                return False

            t_wwpn = map_info["t_wwpn"]
            h_wwpn = map_info["h_wwpn"]
            free_lun_id = self._next_free_h_lun_id(t_wwpn, h_wwpn)

            _print("INFO: Mapping LUN %s to %s/%s (host ID %d)" % (name, t_wwpn, h_wwpn, free_lun_id))

            cmd = 'python -c "import libsan.host.lio as lio;'
            cmd += ' lio.lio_fc_lun_map(\\"{}\\",\\"{}\\",\\"{}\\",\\"{}\\",\\"{}\\")"'.format(
                name,
                "fileio",
                t_wwpn,
                h_wwpn,
                free_lun_id,
            )

            # cmd = ("stlio lun_map --bs_type fileio --lun_name %s --t_wwn %s --h_wwn %s --h_lun_id %s" %
            # (name, t_wwpn, h_wwpn, free_lun_id))
            ret, output = self._run(cmd, return_output=True, verbose=False)
            if ret != 0:
                _print('FAIL: "%s" command' % cmd)
                print(output)
                return False

        self.query_all_lun_info(recheck=True)
        return True

    # TODO
    @staticmethod
    def lun_unmap(name, iqn):
        # Unmap a LUN to an IQN

        _print(f"DEBUG: lun_unmap() - TODO!!! {name} {iqn}")
        return False

    def _next_free_h_lun_id(self, t_wwpn, h_wwpn):
        """Check LIO target and see what is the Host LUN ID that we could use"""
        if not t_wwpn or not h_wwpn:
            _print("FAIL: _next_free_h_lun_id() - requires t_wwpn and h_wwpn")
            return None

        good_lun_id = 0
        all_lun_info = self.query_all_lun_info()
        if not all_lun_info:
            # No LUN is mapped, so return 0
            return good_lun_id

        found_lun_ids = []
        for lun in all_lun_info:
            if "map_infos" not in list(all_lun_info[lun].keys()):
                # LUN is not mapped
                continue
            for map_info in all_lun_info[lun]["map_infos"]:
                if map_info["t_wwpn"] == t_wwpn and map_info["h_wwpn"] == h_wwpn:
                    lun_id = map_info["h_lun_id"]
                    lun_id = lun_id.replace("lun", "")
                    found_lun_ids.append(lun_id)

        if not found_lun_ids:
            return good_lun_id

        return libsan.misc.array.lowest_free_number(found_lun_ids)



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.


"""dt.py: Module to run DT util."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import os
import re  # regex
import subprocess
import sys

import libsan.host.linux
from libsan import _print
from libsan.host.cmdline import exists, run

DT_DEFAULT_OPTION = {
    "flags": "direct",
    "oncerr": "abort",
    "iodir": "forward",
    "align": "0",
    "min": "b",
    "max": "256k",
    "pattern": "incr",
    "dispose": "keep",
}

DT_RW_OPTION = DT_DEFAULT_OPTION
DT_RW_OPTION.update({"disable": "eof,pstats"})

DT_INIT_OPTION = DT_DEFAULT_OPTION
DT_INIT_OPTION.update({"disable": "eof,pstats,compare,verify"})

DT_VERIFY_OPTION = DT_DEFAULT_OPTION
DT_VERIFY_OPTION.update({"disable": "eof,pstats"})

default_timeout = "4h"
default_verbose = 0


def _is_device(path):
    path_regex = re.compile("^/dev/")
    m = path_regex.match(path)
    if m:
        return True
    return False


def _has_dt():
    return exists("dt")


def install_dt():
    if _has_dt():
        return True

    if libsan.host.linux.install_package("dt"):
        return True

    _print("FAIL: Could not install dt")
    return False


def dt_stress(
    of, log=None, time=None, thread=None, limit=None, timeout=None, verbose=False, log_file=None, io_options=None
):
    """Execute DT with IO paramters.
    The arguments are:
    \tof            = Output file or path
    \ttime          = how long dt should run
    \tthread        = The number of disk slices to test.
    \tlimit         = The number of bytes to transfer.
    \tverbose       = If should print command output or not. default: False
    \tio_options    = If want to change default io option parameters. Must be as dict format
    """
    if not io_options:
        io_options = DT_RW_OPTION
    if not _has_dt():
        _print("FATAL: dt is not installed")
        return False

    if log:
        pass
    if timeout:
        pass
    if log_file:
        pass

    io_options["slices"] = thread
    if not io_options["slices"]:
        if _is_device(of):
            # We run 16 threads by default if we are running to a device
            io_options["slices"] = "16"
        else:
            io_options["slices"] = "0"
            # if not limit:
            #    _print("FAIL:  dt_stress(): limit_size must be defined when using file")
            #    return False

    if time:
        io_options["runtime"] = time

    if limit:
        io_options["limit"] = limit

    dt_option = ""
    for key in list(io_options.keys()):
        dt_option += f"{key}={io_options[key]} "
    dt_option += "of=%s" % of

    cmd = "dt %s" % dt_option
    # Append time information to command
    date = 'date "+%Y-%m-%d %H:%M:%S"'
    p = subprocess.Popen(date, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    stdout, _ = p.communicate()
    stdout = stdout.decode("ascii", "ignore")
    stdout = stdout.rstrip("\n")
    _print(f"INFO: [{stdout}] DT Running: '{cmd}'...")

    retcode, output = run(cmd, return_output=True, verbose=verbose)
    if retcode != 0:
        _print("FAIL: running DT")
        print(output)
        return False
    _print("INFO: DT executed successfully")
    return True


def dt_stress_background(of, log=None, time=None, thread=None, limit=None, timeout=None, verbose=False, log_file=None):
    """Run DT on background"""
    newpid = os.fork()
    if newpid == 0:
        # Trying to flush stdout to avoid duplicated lines when running hba_test
        sys.stdout.flush()
        rt = dt_stress(
            of, log=log, time=time, thread=thread, limit=limit, timeout=timeout, verbose=verbose, log_file=log_file
        )
        if not rt:
            os._exit(1)
        # using os._exit() as sys.exit trigger exception
        os._exit(0)
        return None
    else:
        sys.stdout.flush()
        _print("INFO: dt_stress_background(): Child thread %d is running DT Stress" % newpid)
        return newpid


def dt_init_data(of, log=None, limit=None, timeout=None, verbose=False):
    """Usage

        dt_init_data(output_file)
        dt_init_data(of=output_file)
        dt_init_data(of=output_file,log=log)
        dt_init_data(of=output_file,limit=limit_size)
        dt_init_data(of=output_file,timeout=time_out)
        dt_init_data(of=output_file,verbose=flag_verbose)
    Purpose
        Using dt incr patter to write data into output_file. Normally pair with
        dt_verify_data().
        time_out default value is 4h which means I/O could only run 4 hours.
        Set it to 0 if you don't want timeout.
    Parameter
        output_file        # DT I/O output, like "/dev/sda" or "/var/tmp/dt_test.img"
        log                # the sting which hold all log of dt.
        time_in_sec        # time for seconds.
        thread_num         # count of dt thread for I/O stress.
        limit_size         # limit I/O size. Mush have for filesystem stress.
        time_out           # the maximum time dt command can run.
        flag_verbose       # whether dt command output to stdout. default is 0
    Returns
        1
            or
        undef
    Exceptions
        1. no dt installed.
    """
    if not _has_dt():
        _print("FATAL: dt_init_data() - dt is not installed")
        return False

    if not _is_device(of) and not limit:
        _print("FAIL:  dt_init_data(): limit_size must be defined when using file")
        return False

    if log:
        pass

    if timeout:
        pass

    dt_option = ""
    for key in list(DT_INIT_OPTION.keys()):
        dt_option += f"{key}={DT_INIT_OPTION[key]} "

    dt_option += "of=%s" % of

    if limit:
        dt_option += " limit=%s" % limit

    cmd = "dt %s" % dt_option
    # Append time information to command
    date = 'date "+%Y-%m-%d %H:%M:%S"'
    p = subprocess.Popen(date, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    stdout, _ = p.communicate()
    stdout = stdout.decode("ascii", "ignore")
    stdout = stdout.rstrip("\n")
    _print(f"INFO: [{stdout}] DT Running: '{cmd}'...")

    retcode, output = run(cmd, return_output=True, verbose=verbose)
    if retcode != 0:
        _print("FAIL: running DT")
        print(output)
        return False
    _print("INFO: DT initialized data successfully")
    return True


def dt_verify_data(in_file, log=None, limit=None, timeout=None, verbose=False):
    """Usage

        dt_verify_data(input_file)
        dt_verify_data(in_file=output_file)
        dt_verify_data(in_file=output_file,log=log)
        dt_verify_data(in_file=output_file,limit=limit_size)
        dt_verify_data(in_file=output_file,timeout=time_out)
        dt_verify_data(in_file=output_file,verbose=flag_verbose)
    Purpose
        Using dt incr patter to write data into input_file. Normally pair with
        dt_verify_data().
        $time_out default value is 4h which means I/O could only run 4 hours.
        Set it to 0 if you don't want timeout.
    Parameter
        input_file         # DT I/O input, like "/dev/sda" or "/var/tmp/dt_test.img"
        log                # the sting which hold all log of dt.
        time               # any format of time: '30m' or '60s'
        time_in_sec        # time for seconds.
        thread_num         # count of dt thread for I/O stress.
        limit_size         # limit I/O size. Mush have for filesystem stress.
        time_out           # the maxmum time dt command can run.
        flag_verbose       # whether dt command input to stdout. default is 0
    Returns
        1
            or
        undef
    Exceptions
        1. no dt installed.
    """
    if not _has_dt():
        _print("FATAL: dt_init_data() - dt is not installed")
        return False

    if not _is_device(in_file) and not limit:
        _print("FAIL:  dt_init_data(): limit_size must be defined when using file")
        return False

    if log:
        pass

    if timeout:
        pass

    dt_option = ""
    for key in list(DT_INIT_OPTION.keys()):
        dt_option += f"{key}={DT_VERIFY_OPTION[key]} "
    dt_option += "if=%s" % in_file

    if limit:
        dt_option += " limit=%s" % limit

    cmd = "dt %s" % dt_option
    # Append time information to command
    date = 'date "+%Y-%m-%d %H:%M:%S"'
    p = subprocess.Popen(date, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    stdout, _ = p.communicate()
    stdout = stdout.decode("ascii", "ignore")
    stdout = stdout.rstrip("\n")
    _print(f"INFO: [{stdout}] DT Running: '{cmd}'...")

    retcode, output = run(cmd, return_output=True, verbose=verbose)
    if retcode != 0:
        _print("FAIL: running DT")
        print(output)
        return False
    _print("INFO: DT verified data successfully")
    return True



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.


"""fcoe.py: Module to manipulate FCoE devices."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import os.path
import re  # regex

import libsan.host.linux
import libsan.host.net
from libsan import _print
from libsan.host.cmdline import run

supported_soft_fcoe_drivers = ["ixgbe", "bnx2x"]


def configure_intel_fcoe(nic, mode):
    if not libsan.host.linux.service_start("lldpad"):
        return False

    if not libsan.host.net.iface_down(nic):
        return False

    if not libsan.host.net.iface_up(nic):
        return False

    libsan.host.linux.sleep(5)

    retcode, output = run("dcbtool sc %s dcb on" % nic, return_output=True)
    if retcode != 0:
        print(output)
        return False

    libsan.host.linux.sleep(5)

    retcode, output = run("dcbtool sc %s pfc e:1 a:1 w:1" % nic, return_output=True)
    if retcode != 0:
        print(output)
        return False

    retcode, output = run("dcbtool sc %s app:fcoe e:1 a:1 w:1" % nic, return_output=True)
    if retcode != 0:
        print(output)
        return False

    retcode, output = run("cp -f /etc/fcoe/cfg-ethx /etc/fcoe/cfg-%s" % nic, return_output=True)
    if retcode != 0:
        print(output)
        return False

    retcode, output = run(rf"sed -i -e 's/\(MODE=.*\)$/MODE=\"{mode}\"/' /etc/fcoe/cfg-{nic}", return_output=True)
    if retcode != 0:
        print(output)
        return False

    if not libsan.host.linux.service_restart("lldpad"):
        return False

    if not libsan.host.linux.service_restart("fcoe"):
        return False

    if not libsan.host.linux.service_enable("lldpad"):
        return False

    if not libsan.host.linux.service_enable("fcoe"):
        return False

    return True


def configure_bnx2fc_fcoe(nic, mode):
    if not libsan.host.net.iface_up(nic):
        return False

    retcode, output = run("cp -f /etc/fcoe/cfg-ethx /etc/fcoe/cfg-%s" % nic, return_output=True)
    if retcode != 0:
        print(output)
        return False

    retcode, output = run(
        r"sed -i -e 's/\(DCB_REQUIRED=.*\)$/DCB_REQUIRED=\"no\"/' /etc/fcoe/cfg-%s" % nic, return_output=True
    )
    if retcode != 0:
        print(output)
        return False

    retcode, output = run(rf"sed -i -e 's/\(MODE=.*\)$/MODE=\"{mode}\"/' /etc/fcoe/cfg-{nic}", return_output=True)
    if retcode != 0:
        print(output)
        return False

    if not libsan.host.linux.service_restart("lldpad"):
        return False

    if not libsan.host.linux.service_restart("fcoe"):
        return False

    if not libsan.host.linux.service_enable("lldpad"):
        return False

    if not libsan.host.linux.service_enable("fcoe"):
        return False

    return True


def enable_fcoe_on_nic(nic, mode="fabric"):
    if not nic:
        return False

    driver = libsan.host.net.driver_of_nic(nic)
    if not driver:
        _print("FAIL: Could not find driver for %s" % nic)
        return False

    if driver not in supported_soft_fcoe_drivers:
        _print(f"FAIL: NIC {nic} via {driver} is not supported soft FCoE")
        return False

    if libsan.host.linux.is_installed("NetworkManager"):
        conn = libsan.host.net.nm_get_conn(nic)
        libsan.host.net.nm_conn_mod(conn, "connection.autoconnect", "yes")
        libsan.host.net.nm_conn_reload()
        libsan.host.net.nm_conn_up(conn)
    else:
        print("WARN: NetworkManager not found, trying ifcfg..")
        libsan.host.net.set_ifcfg(nic, {"ONBOOT": "yes"})

    configure_fcoe = None
    if driver == "ixgbe":
        configure_fcoe = configure_intel_fcoe
    if driver == "bnx2x":
        configure_fcoe = configure_bnx2fc_fcoe

    if not configure_fcoe(nic, mode):
        print("FAIL: couldn't configure FCoE on %s" % nic)
        return False

    max_wait_time = 180  # wait for maximum 3 minutes
    _print("INFO: Waiting FCoE session with timeout %d seconds" % max_wait_time)
    while max_wait_time >= 0:
        libsan.host.linux.sleep(1)
        max_wait_time -= 1
        # need to query fcoeadmin
        fcoe_dict = query_fcoeadm_i()
        if not fcoe_dict:
            _print("INFO: No FCoE session found, will keep waiting timeout %s seconds" % max_wait_time)
            continue
        for host in list(fcoe_dict.keys()):
            if fcoe_dict[host]["phy_nic"] == nic:
                _print("INFO: FCoE session created for {} as SCSI Host: {}".format(nic, fcoe_dict[host]["scsi_host"]))
                return fcoe_dict
        _print("INFO: No FCoE session created yet for NIC %s, will keep waiting. timeout %d" % (nic, max_wait_time))
    run("fcoeadm -i")
    _print("FAIL: No FCoE session created for NIC %s" % nic)
    return None


def query_fcoeadm_i():
    cmd = "fcoeadm -i"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        if retcode == 2:
            _print("INFO: No FCoE interface is configured")
            return None
        else:
            _print("FAIL: running %s" % cmd)
            print(output)
            return None

    lines = output.split("\n")
    driver_regex = re.compile(r".*Driver:\s+(\S+)\s+(\S+)")
    nic_regex = re.compile(r".*Symbolic Name:\s+(\S+).* over (\S+)")
    scsi_regex = re.compile(r".*OS Device Name: +host(\d+)")
    node_regex = re.compile(r".*Node Name:\s+(\S+)")
    port_regex = re.compile(r".*Port Name:\s+(\S+)")
    fabric_regex = re.compile(r".*Fabric Name:\s+(\S+).*")
    speed_regex = re.compile(r".*  Speed:\s+(.*)")
    supported_speed_regex = re.compile(r".*Supported Speed:\s+(.*)")
    maxframesize_regex = re.compile(r".*MaxFrameSize:\s+(.*)")
    fcid_regex = re.compile(r".*FC-ID \(Port ID\):\s+(\S+)")
    state_regex = re.compile(r".*State:\s+(\S+)")

    fcoeadm_dict = {}
    fcoe_driver = None
    fcoe_driver_version = None
    for line in lines:
        m = driver_regex.match(line)
        if m:
            # this information can be used for more than 1 port
            fcoe_driver = m.group(1)
            fcoe_driver_version = m.group(2)
            continue

        m = nic_regex.match(line)
        if m:
            info_dict = {}
            info_dict["driver"] = fcoe_driver
            info_dict["driver_version"] = fcoe_driver_version
            info_dict["nic"] = m.group(2)
            info_dict["nic_driver"] = libsan.host.net.driver_of_nic(m.group(2))
            info_dict["phy_nic"] = libsan.host.net.phy_nic_of(m.group(2))
            continue

        m = node_regex.match(line)
        if m:
            info_dict["node_name"] = m.group(1)
            continue

        m = port_regex.match(line)
        if m:
            info_dict["port_name"] = m.group(1)
            continue

        m = fabric_regex.match(line)
        if m:
            info_dict["fabric_name"] = m.group(1)
            continue

        m = speed_regex.match(line)
        if m:
            info_dict["speed"] = m.group(1)
            continue

        m = supported_speed_regex.match(line)
        if m:
            info_dict["supported_speed"] = m.group(1)
            continue

        m = maxframesize_regex.match(line)
        if m:
            info_dict["max_frame_size"] = m.group(1)
            continue

        m = fcid_regex.match(line)
        if m:
            info_dict["fcid"] = m.group(1)
            continue

        m = state_regex.match(line)
        if m:
            info_dict["state"] = m.group(1)
            continue

        m = scsi_regex.match(line)
        if m:
            info_dict["scsi_host"] = "host" + m.group(1)
            info_dict["scsi_host_id"] = m.group(1)
            fcoeadm_dict[info_dict["scsi_host_id"]] = info_dict

    if not fcoeadm_dict:
        return None
    return fcoeadm_dict


def setup_soft_fcoe(mode="fabric"):
    """Setup soft FCoE initiator. It supports ixgbe and bnx2x drivers
    The arguments are:
    \tmode:                         The mode defaults to fabric but this option allows the selection of vn2vn mode
    Returns:
    \tTrue:                         If sessions are established
    \tFalse:                        If there was some problem
    """

    libsan.host.linux.install_package("fcoe-utils")

    nic_drv_dict = libsan.host.net.nic_2_driver()
    if not nic_drv_dict:
        _print("FAIL: No NIC found on this server, cannot enable soft FCoE")
        return False

    # print nic_drv_dict
    fcoe_nics = []
    for nic in list(nic_drv_dict.keys()):
        if nic_drv_dict[nic] in supported_soft_fcoe_drivers:
            # _print("INFO: Need to configure %s" % nic)
            fcoe_nics.append(nic)
    if not fcoe_nics:
        _print("INFO: Server has no supported soft FCoE adapter")
        print(nic_drv_dict)
        return False

    fcoeadm_dict = query_fcoeadm_i()
    nic_need_setup = []
    if not fcoeadm_dict:
        nic_need_setup = fcoe_nics
    else:
        # Check if NIC is not already configured
        enabled_nics = []
        for host in list(fcoeadm_dict.keys()):
            enabled_nics.append(fcoeadm_dict[host]["phy_nic"])

        for nic in fcoe_nics:
            if nic not in enabled_nics:
                nic_need_setup.append(nic)
        if not nic_need_setup:
            _print("INFO: All NICs already have FCoE session running:")
            print(fcoe_nics)
            return True

    _print("INFO: Going to enable FCoE on these NICs:")
    for nic in nic_need_setup:
        print(f"\t{nic} ({nic_drv_dict[nic]})")

    error = 0
    for nic in nic_need_setup:
        if not enable_fcoe_on_nic(nic, mode=mode):
            _print("FAIL: Could not enable FCoE on %s" % nic)
            error += 1

    if error:
        run("fcoeadm -i")
        run("ip a")
        return False

    # Wait for server to detect FCoE devices
    _print("INFO: Waiting 120s for devices to be created")
    libsan.host.linux.sleep(120)
    return True


def unconfigure_soft_fcoe():
    """Unconfigure FCoE initiator.
     * unload the related modules
     * stop the fcoe/lldpad daemon
     * chkconfig fcoe/lldpad off
     * delete the /etc/fcoe/cfg-xxx files
     * load the modules

    Returns:
    \tTrue:                         if unconfigure is checked on fcoe initiator
    \tFalse:                        if configure is checked on fcoe initiator
    """

    ret = True
    driver = ""

    # lldpad will still be activated automatically by lldpad.socket
    run("service lldpad stop")
    run("service fcoe stop")

    if run('service fcoe status|grep "Active: active" ') == 0:
        ret = False
        _print("FAIL: Could not stop service of fcoe")

    if run("chkconfig lldpad off") != 0:
        ret = False
        _print("FAIL: Could not disable lldpad on chkconfig")

    if run("chkconfig fcoe off") != 0:
        ret = False
        _print("FAIL: Could not disable fcoe on chkconfig")

    nic_drv_dict = libsan.host.net.nic_2_driver()
    if nic_drv_dict:
        for nic in list(nic_drv_dict.keys()):
            if nic_drv_dict[nic] in supported_soft_fcoe_drivers:
                driver = nic_drv_dict[nic]
                con_path = "/etc/fcoe/cfg-%s" % nic

                if os.path.isfile(con_path) and run("rm -rf %s" % con_path) != 0:
                    ret = False
                    _print("FAIL: Could not delete %s" % con_path)
        if driver:
            if run("modprobe -r %s" % driver) != 0:
                ret = False
                _print("FAIL: Could not unload module %s" % driver)

            if run("modprobe %s" % driver) != 0:
                ret = False
                _print("FAIL: Could not load module %s" % driver)

    return ret


def get_sw_fcoe_nics_n_driver():
    """Get the FCoE nics and its' drivers

    Returns:
    \tstore_dict:                         the dict of store the nics and drivers
    """

    store_dict = {}
    nic_drv_dict = libsan.host.net.nic_2_driver()

    if nic_drv_dict:
        for nic in list(nic_drv_dict.keys()):
            if nic_drv_dict[nic] in supported_soft_fcoe_drivers:
                driver = nic_drv_dict[nic]
                store_dict[nic] = driver

    return store_dict



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""fos.py: Module to handle commands on Brocade FOS Switch."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import re  # regex
import time

import libsan.host.fc
import libsan.host.linux
import libsan.host.ssh
from libsan import _print


def parse_ret(data):
    # remove command on 1st line and last line with '> '
    return data.split("\r\n")[1:-1]


class fos:  # noqaN801
    """Class to manage Brocade FOS switches"""

    host = None
    user = None
    passwd = None

    port_info_dict = None

    def __init__(self, hostname, username, password, timeout=None):
        self.host = hostname
        self.user = username
        self.passwd = password
        self.timeout = timeout
        if self.timeout:
            self.timeout = float(self.timeout)

    def _run(self, cmd, return_output=False, verbose=True):
        session = libsan.host.ssh.connect(self.host, user=self.user, passwd=self.passwd)
        if not session:
            _print("FAIL: Could not connect to switch (%s)" % self.host)
            if return_output:
                return 127, None
            return 127
        ret = libsan.host.ssh.run_cmd(
            session,
            cmd=cmd,
            return_output=return_output,
            verbose=verbose,
            timeout=self.timeout,
            invoke_shell=True,
            expect="> ",
        )

        libsan.host.ssh.disconnect(session)

        if return_output:
            ret_code, output = ret
            output = "\n".join(parse_ret(output))
            # print output
            return ret_code, output

        return ret

    def get_version(self):
        """Return FW version of the switch"""
        cmd = "version"
        _, output = self._run(cmd, return_output=True, verbose=False)
        version_regex = re.compile(r".*Fabric OS:\s+(\S+)", re.MULTILINE)
        v = version_regex.search(output)
        if v:
            return "%s" % (v.group(1))
        return None

    @staticmethod
    def capability():
        """Indicates supported operation on switch"""
        cap_dict = {}
        cap_dict["link_down"] = True
        cap_dict["link_up"] = True
        cap_dict["link_oscillate"] = True
        cap_dict["switch_reboot"] = False
        cap_dict["fc_switch"] = True
        cap_dict["eth_switch"] = False
        cap_dict["rcsn_enable"] = True
        cap_dict["rcsn_disable"] = True
        return cap_dict

    def show_interfaces(self):
        cmd = "switchshow"  # % option
        return self._run(cmd)

    def link_trigger(self, action=None, port_id=None, check_state=True):
        if not action or not port_id:
            _print("FAIL: link_trigger needs 'action' and 'port_id' parameters")
            return False

        if action == "UP":
            cmds = "portenable %s" % port_id
        elif action == "DOWN":
            cmds = "portdisable %s" % port_id
        else:
            _print("FAIL: link_trigger unsupported action (%s)" % action)
            return False
        ret = self._run(cmds, verbose=True)
        if ret != 0:
            _print("FAIL: link_trigger command failed")
            return False

        # We do not want make sure the state is correct
        if not check_state:
            return True

        # wait up to 10s to detect link change
        cnt = 0
        max_attempts = 5
        while cnt < max_attempts:
            cnt += 1
            time.sleep(2)

            p_info_dict = self.query_port_info(recheck=True)
            if self.port_state(port_id) == action:
                _print(f"INFO: Successfully brought port {port_id} {action}")
                return True

            if cnt == max_attempts:
                _print(f"FAIL: Failed to bring port {port_id} {action}. Current state: {p_info_dict[port_id]}")
                return False

        return False

    def port_state(self, port_id):
        if not port_id:
            _print("FAIL: port_state requires 'port_id' as parameter")
            return None

        p_info_dict = self.query_port_info()
        if not p_info_dict:
            _print("FAIL: Could not query port info")
            return None

        if port_id not in list(p_info_dict.keys()):
            _print(f"FAIL: port {port_id} was not found on switch {self.host}")
            print(p_info_dict)
            return None

        if "state" not in list(p_info_dict[port_id].keys()):
            _print("FAIL: Could not find port state for %s" % port_id)
            print(p_info_dict)
            return None

        return p_info_dict[port_id]["state"]

    def port_connect(self, port_id, check_state=True):
        return self.link_trigger(action="UP", port_id=port_id, check_state=check_state)

    def port_disconnect(self, port_id, check_state=True):
        return self.link_trigger(action="DOWN", port_id=port_id, check_state=check_state)

    def port_oscillate(self, port_id, min_uptime, max_uptime, min_downtime, max_downtime, count):
        if (
            not port_id
            or min_uptime is None
            or max_uptime is None
            or min_downtime is None
            or max_downtime is None
            or count is None
        ):
            _print(
                "FAIL: port_oscillate() - needs 'port_id', 'min_uptime', "
                "'max_uptime', 'min_downtime', 'max_downtime' and 'count' parameters"
            )
            return False

        for cnt in range(1, count + 1):
            _print("INFO: Interaction %d/%d" % (cnt, count))
            self.port_disconnect(port_id, check_state=False)
            increment = (max_downtime - min_downtime) / count
            sleep_time = min_downtime + ((cnt - 1) * increment)
            _print("INFO: Going to sleep for %s with link down" % sleep_time)
            libsan.host.linux.sleep(sleep_time)
            self.port_connect(port_id, check_state=False)
            increment = (max_uptime - min_uptime) / count
            sleep_time = min_uptime + ((cnt - 1) * increment)
            _print("INFO: Going to sleep for %s with link up" % sleep_time)
            libsan.host.linux.sleep(sleep_time)

        return True

    def query_port_info(self, recheck=False):
        """Query all interfaces on switch and store its information on a dict
        If port is 'online' we set it as UP
        Otherwise we set it as DOWN
        """

        # If we do not need to recheck and port info exist we return it
        if self.port_info_dict and not recheck:
            return self.port_info_dict

        cmd = "switchshow"
        ret, output = self._run(cmd, return_output=True, verbose=False)
        if ret != 0:
            _print("FAIL: switchshow command")
            print(output)
            return None
        # Index Port Address Media Speed State     Proto
        #    0   0   010000   id    N8   Online      FC  F-Port  50:0a:09:81:99:1b:8d:c5
        info_regex = re.compile(r"\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s?(.*)")
        # extra parameters, not always present
        # for example: Disabled or F-Port  50:0a:09:82:89:1b:8d:c5 or (No POD License) Disabled
        info_extra_regex = re.compile(r"(.*)\s(\S+)")
        lines = output.split("\n")

        self.port_info_dict = {}
        # Process port status
        for line in lines:
            m = info_regex.match(line)
            if m:
                # print line
                p_dict = {}
                port_id = m.group(1)
                p_dict["wwpn"] = None
                if m.group(8):
                    e = info_extra_regex.match(m.group(8))
                    if e and libsan.host.fc.is_wwn(e.group(2)):
                        p_dict["wwpn"] = e.group(2)

                state = m.group(6)
                if state == "No_Module":
                    # Skip
                    continue
                state = "UP" if state == "Online" else "DOWN"
                p_dict["state"] = state
                self.port_info_dict[port_id] = p_dict

        # Get the wwpns connected to each port
        # We might need to do this, because some ports have more than 1 WWPN connected to it (NPIV...)
        process_wwpns = False
        for port_id in self.port_info_dict:
            cmd = "portshow %s" % port_id
            ret, output = self._run(cmd, return_output=True, verbose=False)
            if ret != 0:
                _print("FAIL: %s" % cmd)
                return None
            lines = output.split("\n")
            begin_regex = re.compile(r"portWwn of device\(s\) connected:")
            end_regex = re.compile(r"Distance:")
            wwpn_regex = re.compile(r"\s+((?:[0-9a-f]{2}:){7}[0-9a-f]{2})")
            for line in lines:
                if begin_regex.match(line):
                    process_wwpns = True
                    continue
                if end_regex.match(line):
                    process_wwpns = False
                    continue
                # There could have more then 1 wwpn, for example NPIVs
                #       #For now we just store the last entry...
                if process_wwpns:
                    m = wwpn_regex.search(line)
                    if m:
                        self.port_info_dict[port_id]["wwpn"] = m.group(1)

        return self.port_info_dict

    def wwpn_2_port_id(self):
        """Query all interfaces that are connected
        And store as key the WWPN and the Port ID as its value
        """
        port_dict = self.query_port_info()

        if not port_dict:
            _print("FAIL: Could not query port info")
            return None

        wwn_dict = {}
        for port_id in list(port_dict.keys()):
            if "wwpn" in list(port_dict[port_id].keys()):
                wwn_dict[port_id] = port_dict[port_id]["wwpn"]
        return wwn_dict

    @staticmethod
    def mac_2_port_id():
        """Query all interfaces that are connected
        And store as key the WWPN and the Port ID as its value
        """
        # TODO
        return

    def rcsn_trigger(self, action=None, port_id=None):
        if not action or not port_id:
            _print("FAIL: rcsn_trigger needs 'action' and 'port_id' parameters")
            return False

        cmds = None
        # To stop RCSN messages to being sent, we need to enable RCSN suppresion
        if action == "ENABLE":
            cmds = ["portcfg rscnsupr %s --disable" % port_id]
        elif action == "DISABLE":
            cmds = ["portcfg rscnsupr %s --enable" % port_id]
        else:
            _print("FAIL: rcsn_trigger unsupported action (%s)" % action)
        ret = self._run(cmds, verbose=True)
        if ret != 0:
            _print("FAIL: rcsn_trigger command failed")
            return False
        return True



#!/usr/bin/python

# Copyright (C) 2021 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.


import unittest
from unittest.mock import call, patch

from libsan.host import fcoe
from libsan.host.cmdline import run

nic_2_driver = {"test_nic1": "ixgbe", "test_nic2": "bnx2x"}

fake_fcoeadm1 = {"host1": {"phy_nic": "test_nic1", "scsi_host": "host1"}}
fake_fcoeadm2 = {
    "host1": {"phy_nic": "test_nic1", "scsi_host": "host1"},
    "host2": {"phy_nic": "test_nic2", "scsi_host": "host2"},
}
# simulate first just one connection, than both
query_fcoeadm_i_outputs = [{}, {}, fake_fcoeadm1, fake_fcoeadm2]


def driver_of_nic(nic):
    return nic_2_driver[nic]


def _run(cmd, return_output=True, verbose=False):
    # To configure soft fcoe we need some tools to be available
    # We need to make sure setup_soft_fcoe would install them
    tool = cmd.split()[0]
    ret, _ = run(f"command -v {tool}", return_output=return_output, verbose=verbose)
    if ret != 0:
        return ret, f"Required command {tool} not installed"
    return 0, ""


class TestFCoE(unittest.TestCase):
    @patch("libsan.host.fcoe.run")
    @patch("libsan.host.linux.sleep")
    @patch("libsan.host.net.iface_up")
    @patch("libsan.host.net.iface_down")
    @patch("libsan.host.linux.service_enable")
    @patch("libsan.host.linux.service_restart")
    @patch("libsan.host.linux.service_start")
    def test_configure_intel_fcoe(
        self,
        service_start_func,
        service_restart_func,
        service_enable_func,
        iface_down_func,
        iface_up_func,
        sleep_func,
        run_func,
    ):
        service_start_func.return_value = True
        service_restart_func.return_value = True
        service_enable_func.return_value = True
        iface_down_func.return_value = True
        iface_up_func.return_value = True
        sleep_func.return_value = True
        run_func.return_value = 0, ""

        assert fcoe.configure_intel_fcoe("test_nic1", "fabric")

        assert run_func.call_count == 5
        run_calls = [
            call("dcbtool sc test_nic1 dcb on", return_output=True),
            call("dcbtool sc test_nic1 pfc e:1 a:1 w:1", return_output=True),
            call("dcbtool sc test_nic1 app:fcoe e:1 a:1 w:1", return_output=True),
            call("cp -f /etc/fcoe/cfg-ethx /etc/fcoe/cfg-test_nic1", return_output=True),
            call("sed -i -e 's/\\(MODE=.*\\)$/MODE=\\\"fabric\\\"/' /etc/fcoe/cfg-test_nic1", return_output=True),
        ]
        run_func.assert_has_calls(run_calls)

    @patch("libsan.host.fcoe.run")
    @patch("libsan.host.net.iface_up")
    @patch("libsan.host.linux.service_enable")
    @patch("libsan.host.linux.service_restart")
    def test_configure_bnx2fc_fcoe(self, service_restart_func, service_enable_func, iface_up_func, run_func):
        service_restart_func.return_value = True
        service_enable_func.return_value = True
        iface_up_func.return_value = True
        run_func.return_value = 0, ""

        assert fcoe.configure_bnx2fc_fcoe("test_nic2", "fabric")

        assert run_func.call_count == 3
        run_calls = [
            call("cp -f /etc/fcoe/cfg-ethx /etc/fcoe/cfg-test_nic2", return_output=True),
            call(
                "sed -i -e 's/\\(DCB_REQUIRED=.*\\)$/DCB_REQUIRED=\\\"no\\\"/' /etc/fcoe/cfg-test_nic2",
                return_output=True,
            ),
            call("sed -i -e 's/\\(MODE=.*\\)$/MODE=\\\"fabric\\\"/' /etc/fcoe/cfg-test_nic2", return_output=True),
        ]
        # print(run_func.call_args_list)
        run_func.assert_has_calls(run_calls)

    @patch("libsan.host.fcoe.query_fcoeadm_i")
    @patch("libsan.host.fcoe.configure_bnx2fc_fcoe")
    @patch("libsan.host.fcoe.configure_intel_fcoe")
    @patch("libsan.host.net.nm_conn_up")
    @patch("libsan.host.net.nm_conn_reload")
    @patch("libsan.host.net.nm_conn_mod")
    @patch("libsan.host.net.nm_get_conn")
    @patch("libsan.host.linux.is_installed")
    @patch("libsan.host.net.driver_of_nic")
    def test_enable_fcoe_on_nic(
        self,
        driver_of_nic_func,
        is_installed_func,
        nm_get_conn_func,
        nm_conn_mod_func,
        nm_conn_reload_func,
        nm_conn_up_func,
        configure_intel_fcoe_func,
        configure_bnx2fc_fcoe_func,
        query_fcoeadm_i_func,
    ):
        driver_of_nic_func.side_effect = driver_of_nic
        is_installed_func.return_value = True
        nm_get_conn_func.return_value = "test_conn"
        nm_conn_mod_func.return_value = True
        nm_conn_reload_func.return_value = True
        nm_conn_up_func.return_value = True
        configure_intel_fcoe_func.return_value = True
        configure_bnx2fc_fcoe_func.return_value = True
        query_fcoeadm_i_func.side_effect = query_fcoeadm_i_outputs

        assert fcoe.enable_fcoe_on_nic("test_nic1")
        assert fcoe.enable_fcoe_on_nic("test_nic2")



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""iscsi.py: Module to manipulate iSCSI devices."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import re  # regex

import libsan.host.linux
import libsan.host.mp
import libsan.host.net
import libsan.host.scsi
from libsan import _print
from libsan.host.cmdline import run

# used to match regex for each session information that we support
supported_discovery_info = {
    "address": r".*DiscoveryAddress: (\S+)",
    "target": r".*Target: (\S+)",
    "portal": r".*Portal: (\S+):(\S+),(\S+)",
    "iface": r".*Iface Name: (\S+)",
}

# used to match regex for each session information that we support
supported_session_info = {
    "t_iqn": r".*Target: (\S+)",
    "h_iqn": r".*Iface Initiatorname: (\S+)",
    "iface": r".*Iface Name: (\S+)",
    "transport": r".*Iface Transport: (\S+)",
    "iface_ip": r".*Iface IPaddress: (\S+)",
    "mac": r".*Iface HWaddress: (\S+)",
    "sid": r".*SID: (\S+)",
    "host": r".*Host Number: (\S+).*State: (\S+)",  # e.g. Host Number: 6	State: running
    "disks": r".*Attached scsi disk (\S+).*State: (\S+)",
    # eg. Attached scsi disk sdb		State: running
    "target_ip": r".*Current Portal: (\S+):[0-9]+,",
    "persist_ip": r".*Persistent Portal: (\S+):[0-9]+,",
    # negotiated parameters
    "header_digest": r".*HeaderDigest: (\S+)",
    "data_digest": r".*DataDigest: (\S+)",
    "max_recv": r".*MaxRecvDataSegmentLength: (\S+)",
    "max_xmit": r".*MaxXmitDataSegmentLength: (\S+)",
    "first_burst": r".*FirstBurstLength: (\S+)",
    "max_burst": r".*MaxBurstLength: (\S+)",
    "immediate_data": r".*ImmediateData: (\S+)",
    "initial_r2t": r".*InitialR2T: (\S+)",
    "max_outst_r2t": r".*MaxOutstandingR2T: (\S+)",
}

host_path = "/sys/class/iscsi_host/"


def is_iqn(iqn):
    if re.match(r"^iqn\.", iqn):
        return True
    return False


def install():
    """Install iscsiadm tool
    The arguments are:
    \tNone
    Returns:
    \tTrue: If iscsiadm is installed correctly
    \tFalse: If some problem happened
    """
    pack = "iscsi-initiator-utils"
    if not libsan.host.linux.install_package(pack):
        _print("FAIL: Could not install %s" % pack)
        return False

    return True


# Return an array with all iscsi_hosts numbers
def get_iscsi_hosts():
    cmd = "ls " + host_path
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        return None
    # remove 'host' prefix
    output = re.sub("host", "", output)
    host_array = output.split()
    return host_array


# iSCSI discovery ###
def query_discovery():
    """Query all iSCSI targets
    The arguments are:
    \tNone
    Returns:
    \tDict:    Dict with all discovered targets
    \tNone:    If some problem happened
    """
    cmd = "iscsiadm -m discovery -P1"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        # If no target is found iscsiadm returns error code
        return None
    lines = output.split("\n")

    supported_discovery_modes = ["SENDTARGETS", "iSNS", "STATIC", "FIRMWARE"]
    supported_mode_type = {"SENDTARGETS": "sendtargets", "iSNS": "isns"}

    discovery_info_dict = {}
    discovery_address = None
    disc_mode = None
    target_name = None

    for line in lines:
        # print "(%s)" % line
        # Check if it is discovery mode information
        m = re.match("(^.*):", line)
        if m and m.group(1) in supported_discovery_modes:
            disc_mode = m.group(1)
            # We will use DiscoveryAddress as key
            discovery_info_dict[disc_mode] = {}
            discovery_address = None
            continue

        # We will use TargetAddress as key for the target dictionary
        m = re.match(supported_discovery_info["address"], line)
        if m:
            discovery_address = m.group(1)
            if discovery_address not in list(discovery_info_dict[disc_mode].keys()):
                discovery_info_dict[disc_mode][discovery_address] = {}
            disc_addr_regex = re.compile(r"(\S+),(\S+)")
            d = disc_addr_regex.match(discovery_address)
            if d:
                discovery_info_dict[disc_mode][discovery_address]["disc_addr"] = d.group(1)
                discovery_info_dict[disc_mode][discovery_address]["disc_port"] = d.group(2)

            if disc_mode in list(supported_mode_type.keys()):
                discovery_info_dict[disc_mode][discovery_address]["mode"] = supported_mode_type[disc_mode]
            continue

        m = re.match(supported_discovery_info["target"], line)
        if m:
            # FIRMWARE discovery might not use discovery address
            if not discovery_address:
                discovery_address = "NotSet"
                discovery_info_dict[disc_mode][discovery_address] = {}

            target_name = m.group(1)
            if "targets" not in list(discovery_info_dict[disc_mode][discovery_address].keys()):
                discovery_info_dict[disc_mode][discovery_address]["targets"] = {}
            discovery_info_dict[disc_mode][discovery_address]["targets"][target_name] = {}
            continue

        m = re.match(supported_discovery_info["portal"], line)
        if m:
            discovery_info_dict[disc_mode][discovery_address]["targets"][target_name]["portal"] = {}
            discovery_info_dict[disc_mode][discovery_address]["targets"][target_name]["portal"]["address"] = m.group(1)
            discovery_info_dict[disc_mode][discovery_address]["targets"][target_name]["portal"]["port"] = m.group(2)
            continue

        m = re.match(supported_discovery_info["iface"], line)
        if m:
            iface = m.group(1)
            if "iface" not in list(discovery_info_dict[disc_mode][discovery_address]["targets"][target_name].keys()):
                discovery_info_dict[disc_mode][discovery_address]["targets"][target_name]["iface"] = []
            discovery_info_dict[disc_mode][discovery_address]["targets"][target_name]["iface"].append(iface)
            continue
            # print "Found %s: %s" % (key, m.group(1))

    return discovery_info_dict


def discovery_st(target, ifaces=None, disc_db=False):
    """Discover iSCSI target
    The arguments are:
    \ttarget:   Address of target to be discovered
    \tifaces:   iSCSI interfaces to be used, separated by space (optional)
    \tdisc_db:  To use discoverydb instead of discovery (optional)
    Returns:
    \tTrue:     If it discovered an iSCSI target
    \tFalse:    If some problem happened
    """
    print("INFO: Executing Discovery_ST() with these arges:")
    print("\tTarget: %s" % target)
    if ifaces:
        print("\tIfaces: %s" % ifaces)

    disc_opt = "discovery"
    operation = None

    if disc_db:
        disc_opt = "discoverydb -D"
        operation = "new"

    cmd = f"iscsiadm -m {disc_opt} -p {target}"
    if operation:
        cmd += " -o %s" % operation

    if ifaces:
        if ("bnx2i" in ifaces or "qedi" in ifaces) and libsan.host.linux.service_status("iscsiuio", verbose=False) != 0:
            libsan.host.linux.service_enable("iscsiuio", now=True)
        interfaces = ifaces.split(" ")
        for interface in interfaces:
            cmd += " -I %s" % interface
    cmd += " -t st"
    retries = 0
    retcode, data = run(cmd, return_output=True, verbose=True)
    while retcode == 0 and "(err 29)" in data and retries < 5:
        retcode, data = run(cmd, return_output=True, verbose=True)
        retries += 1
    if retcode != 0 or retries == 5:
        _print("FAIL: Could not discover iSCSI target")
        return False
    return True


def is_target_discovered(t_iqn):
    """Check if an iSCSI target is already discovered
    The arguments are:
    \tiSCSI Target:   iQN of iSCSI target
    Returns:
    \tTrue:     If target is discovered
    \tFalse:    If was not found
    """
    if not t_iqn:
        _print("FAIL: is_target_discovered() - requires target iqn as parameter")

    disc_dict = query_discovery()
    if not disc_dict:
        return False

    for disc_type in list(disc_dict.keys()):
        for disc_addr in list(disc_dict[disc_type].keys()):
            if "targets" not in list(disc_dict[disc_type][disc_addr].keys()):
                continue
            if t_iqn in list(disc_dict[disc_type][disc_addr]["targets"].keys()):
                # Target is already discovered we do not need to do anything
                return True
    return False


def get_disc_ifaces_of_t_iqn(t_iqn):
    """From given target IQN, return the interfaces that discovered it
    The arguments are:
    \tiSCSI Target:   iQN of iSCSI target
    Returns:
    \tList ifaces:     Discovered interfaces
    \tNone:             If iface was not found
    """

    if not t_iqn:
        _print("FAIL: get_t_iqn_disc_ifaces() - requires target iqn")
        return None

    if not is_target_discovered(t_iqn):
        _print("FAIL: get_t_iqn_disc_ifaces() - target iqn: %s is not discovered" % t_iqn)
        return None

    disc_dict = query_discovery()
    for disc_type in list(disc_dict.keys()):
        for disc_addr in list(disc_dict[disc_type].keys()):
            if "targets" not in list(disc_dict[disc_type][disc_addr].keys()):
                continue
            if t_iqn in list(disc_dict[disc_type][disc_addr]["targets"].keys()):  # noqa: SIM102
                if "iface" in list(disc_dict[disc_type][disc_addr]["targets"][t_iqn].keys()):
                    return disc_dict[disc_type][disc_addr]["targets"][t_iqn]["iface"]
    return None


def delete_discovery_target_portal(portal, port="3260", tp="st"):
    """Delete discovered iSCSI target
    The arguments are:
    \tportal:   Address of target to be discovered
    \tport:     Port of iSCSI target to be deleted
    \ttp:       Discovery type, sendtargets, isns...
    Returns:
    \tTrue:     If deleted discovered iSCSI target
    \tFalse:    If some problem happened
    """
    _print("INFO: Deleting target portal: %s" % portal)
    if libsan.host.net.get_ip_version(portal) == 6:
        # IF IPv6 we need to append "[" and "]" to the address
        portal = "[" + portal + "]"

    cmd = f'iscsiadm -m discoverydb --type {tp} --portal "{portal}:{port}" -o delete'
    retcode, _ = run(cmd, return_output=True, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not delete discover iSCSI target")
        return False
    return True


def clean_up(portal="all"):
    """Remove iSCSI session and discover information for specific target
    The arguments are:
    \ttarget:   Address of target to be removed
    Returns:
    \tTrue:     If iSCSI target is removed
    \tFalse:    If some problem happened
    """

    error = 0
    # TODO: iSCSI boot clean up
    if is_iscsi_boot():
        boot_dev = libsan.host.linux.get_boot_device()
        if not boot_dev:
            _print("FAIL: clean_up() - Could not determine boot device")
            return False

        boot_wwid = libsan.host.linux.get_device_wwid(boot_dev)
        if not boot_wwid:
            _print("FAIL: clean_up() - Could not determine boot WWID for %s" % boot_dev)
            return False

        ses_ids = get_all_session_ids()
        if not ses_ids:
            _print("FAIL: is_iscsi_boot() - It is iSCSI boot, but did not find any session ID")
            return False

        if portal == "all":
            # Logout from all iSCSI session, that do not have boot device
            for ses_id in ses_ids:
                iscsi_wwids = scsi_wwid_of_iscsi_session(sid=ses_id)
                if boot_wwid in iscsi_wwids:
                    _print("INFO: Can't log out of session %s, because it is used for iSCSI boot" % ses_id)
                else:
                    _print("INFO: Logging out of session %s" % ses_id)
                    session_logout(ses_id)
                    # TODO Clean up discovery info
        else:
            # TODO Logout single portal from iSCSI boot
            _print("FAIL: clean_up() - Does not know how to clean up portal %s for iSCSI boot" % portal)
            return False

        return True

    # Not iSCSI boot
    if portal == "all":
        # log out of all iSCSI sessions
        if get_all_session_ids():  # noqa: SIM102
            # There is at least one session
            if not node_logout():
                _print("FAIL: Could not logout from %s iSCSI target" % portal)
                error += 1
    else:
        if not node_logout(portal=portal):
            _print("FAIL: Could not logout from %s iSCSI target" % portal)
            error += 1

    disc_dict = query_discovery()
    # If there is discovery information
    if disc_dict:
        # We will search for this portal on sendtargets and iSNS
        for mode in list(disc_dict.keys()):
            if mode != "SENDTARGETS" and mode != "iSNS":
                # We only delete discover info for st and isns
                continue
            m_dict = disc_dict[mode]
            # Search for all discovered address if they match the one given
            for addr in list(m_dict.keys()):
                d_dict = m_dict[addr]

                disc_addr = d_dict["disc_addr"]
                port = d_dict["disc_port"]
                if disc_addr == portal or portal == "all":  # noqa: SIM102
                    if not delete_discovery_target_portal(disc_addr, port=port, tp=d_dict["mode"]):
                        _print("FAIL: Deleting iSCSI target %s" % d_dict["disc_addr"])
                        error += 1

    if error:
        return False
    return True


# iSCSI session ###
# def query_sessions():
#    #cmd output: tcp: [21] 127.0.0.1:3260,1 iqn.2009-10.com.redhat:storage-1 (non-flash)
#    cmd = "iscsiadm -m session"
#    retcode, output = run(cmd, return_output=True, verbose=False)
#    if (retcode != 0):
#        return None
#    lines = output.split("\n")
#    session_regex = re.compile("(\S+):\s[(\d+)]\s(\S+):(\S+),(\d+),(\S+)")
#    sessions_dict = {}
#    for line in lines:
#        m = session_regex.search(line)
#        if m:
#            sid = m.group(2)
#            ses_dict = {}
#            ses_dict["driver"] = m.group(1)
#            ses_dict["portal"] = m.group(3)
#            ses_dict["portal_port"] = m.group(4)
#            ses_dict["target_iqn"] = m.group(6)
#            sessions[sid] = ses_dict
#    return sessions_dict


def get_all_session_ids():
    cmd = "iscsiadm -m session -P1"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        # _print ("INFO: there is no iSCSI session")
        return None
    lines = output.split("\n")

    session_ids = []

    for line in lines:
        m = re.match(supported_session_info["sid"], line)
        if not m:
            continue
        # print "Found session id: %s" %m.group(1)
        session_ids.append(m.group(1))
    return session_ids


def query_iscsi_session(sid):
    """Query information from a specific iSCSI session
    The arguments are:
    \tsid:      Session ID
    Returns:
    \tDict:     A dictionary with session info
    \tNone:     If some problem happened
    """
    if not sid:
        _print("FAIL: query_iscsi_session() - requires sid as argument")
        return None

    regex_session_scsi_id = "^[ \t]+scsi([0-9]+) Channel ([0-9]+) Id ([0-9])+ Lun: ([0-9]+)$"
    cmd = "iscsiadm -m session -P3 -S -r %s" % sid
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        return None
    lines = output.split("\n")

    session_info_dict = {}
    # dict with disk name and its status
    session_disks_dict = {}
    # store host number and status
    session_host_dict = {}
    for line in lines:
        # print "(%s)" % line

        m = re.match(regex_session_scsi_id, line)
        if m:
            host_id = m.group(1)
            target_id_only = m.group(2)
            bus_id_only = m.group(3)
            lun_id = m.group(4)
            target_id_only = re.sub("^0+(?=.)", "", target_id_only)
            scsi_id = f"{host_id}:{target_id_only}:{bus_id_only}:{lun_id}"

            if "scsi_id_info" not in list(session_info_dict.keys()):
                session_info_dict["scsi_id_info"] = {}
            session_info_dict["scsi_id_info"][scsi_id] = {}
            session_info_dict["scsi_id_info"][scsi_id]["scsi_id"] = scsi_id

        # Could be more than one scsi disk, will add as dict
        m = re.match(supported_session_info["disks"], line)
        if m:
            disk_dict = {}
            # disk_dict["scsi_name"] = m.group(1)
            disk_dict["status"] = m.group(2)
            disk_dict["wwid"] = libsan.host.scsi.wwid_of_disk(m.group(1))
            session_disks_dict[m.group(1)] = disk_dict
            continue

        # Could be more than one scsi disk, will add as dict
        m = re.match(supported_session_info["host"], line)
        if m:
            session_host_dict[m.group(1)] = m.group(2)
            continue
        # Generic search for keys and values
        for key in list(supported_session_info.keys()):
            m = re.match(supported_session_info[key], line)
            if not m:
                continue
            # print "Found %s: %s" % (key, m.group(1))
            session_info_dict[key] = m.group(1)
            if session_info_dict[key] == "<empty>":
                session_info_dict[key] = None
                if key == "mac":  # noqa: SIM102
                    # Try to get based on iface IP address
                    if "iface_ip" in list(session_info_dict.keys()):
                        nic = libsan.host.net.get_nic_of_ip(session_info_dict["iface_ip"])
                        if nic:
                            session_info_dict[key] = libsan.host.net.get_mac_of_nic(nic)
    # added info for the specific session
    session_info_dict["disks"] = session_disks_dict
    session_info_dict["host"] = session_host_dict
    return session_info_dict


def query_all_iscsi_sessions():
    """First we get all iSCSI ids, later on we get the information of each session individually"""

    session_ids = get_all_session_ids()
    if not session_ids:
        return None

    iscsi_sessions = {}
    # Collecting info from each session
    for sid in session_ids:
        session_info_dict = query_iscsi_session(sid)
        iscsi_sessions[sid] = session_info_dict

    # print iscsi_sessions
    return iscsi_sessions


def session_logout(sid=None):
    cmd = "iscsiadm -m session -u"
    if sid:
        cmd += " -r %s" % sid
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        print(output)
        _print("FAIL: session_logout() - Could not logout from session")
        return None
    return True


def get_iscsi_session_by_scsi_id(scsi_id):
    """Return the Session Dict that has the scsi_id"""
    sessions = query_all_iscsi_sessions()
    if not sessions:
        return None

    for ses in sessions:
        if "scsi_id_info" not in list(sessions[ses].keys()):
            continue
        if scsi_id in list(sessions[ses]["scsi_id_info"].keys()):
            return sessions[ses]
    return None


def h_iqn_of_sessions():
    """Usage
        h_iqn_of_sessions()
    Purpose
        Get the Host IQNs of all active iSCSI sessions
    Parameter
        None
    Returns
        List:   h_iqns
            or
        None
    """
    h_iqns = None
    sessions = query_all_iscsi_sessions()
    if not sessions:
        return None

    for key in list(sessions.keys()):
        info = sessions[key]
        if "h_iqn" in list(info.keys()):
            if not h_iqns:
                h_iqns = []
            if info["h_iqn"] not in h_iqns:
                h_iqns.append(info["h_iqn"])
    return h_iqns


def t_iqn_of_sessions():
    """Usage
        t_iqn_of_sessions()
    Purpose
        Get the Target IQNs of all active iSCSI sessions
    Parameter
        None
    Returns
        List:   t_iqns
            or
        None
    """
    t_iqns = None
    sessions = query_all_iscsi_sessions()
    if not sessions:
        return None

    for key in list(sessions.keys()):
        info = sessions[key]
        if "t_iqn" in list(info.keys()):
            if not t_iqns:
                t_iqns = []
            if info["t_iqn"] not in t_iqns:
                t_iqns.append(info["t_iqn"])
    return t_iqns


def mac_of_iscsi_session():
    """Usage
        mac_of_iscsi_session()
    Purpose
        We only check host IQN in active iSCSI session.
    Parameter
        None
    Returns
        List:   macs
            or
        None
    """
    macs = None
    sessions = query_all_iscsi_sessions()
    if not sessions:
        return None

    for key in list(sessions.keys()):
        info = sessions[key]
        if "mac" in list(info.keys()):
            if not macs:
                macs = []
            if info["mac"] != "<empty>" and info["mac"] and info["mac"] not in macs:
                macs.append(info["mac"])
    return macs


def scsi_names_of_iscsi_session(h_iqn=None, t_iqn=None, sid=None):
    """Usage
        scsi_names_of_iscsi_session();
        scsi_names_of_iscsi_session(sid=1);
        scsi_names_of_iscsi_session(h_iqn=h_iqn, t_iqn=t_iqn);
    # we should not support this method since the h_iqn for qla4xxx
    #    scsi_names_of_iscsi_session(t_iqn=t_iqn, h_iqn=h_iqn);
        scsi_names_of_iscsi_session(iface=iface,target_ip=target_ip,;
            t_iqn=t_iqn);
        scsi_names_of_iscsi_session(session_id=session_id);
    Purpose
        Query out all SCSI device names for certain iscsi session.
    Parameter
        h_iqn                  # the IQN used by the host
        t_iqn                  # the IQN used by iscsi target
        sid                    # the iSCSI session ID
    Returns
        scsi_names
            or
        None
    """
    sessions = query_all_iscsi_sessions()
    if not sessions:
        return None

    if sid:
        if sid in list(sessions.keys()) and "disks" in list(sessions[sid].keys()):
            return list(sessions[sid]["disks"].keys())
        return None

    scsi_names = None
    if not h_iqn and not t_iqn:
        for sid in list(sessions.keys()):
            if "disks" in list(sessions[sid].keys()):
                if not scsi_names:
                    scsi_names = []
                scsi_names.extend(list(sessions[sid]["disks"].keys()))
        return scsi_names

    if h_iqn and t_iqn:
        for sid in list(sessions.keys()):
            if sessions[sid]["h_iqn"] == h_iqn and sessions[sid]["t_iqn"] == t_iqn:  # noqa: SIM102
                if "disks" in list(sessions[sid].keys()):
                    if not scsi_names:
                        scsi_names = []
                    scsi_names.extend(list(sessions[sid]["disks"].keys()))
        return scsi_names

    _print("FAIL: scsi_names_of_iscsi_session() - Unsupported parameters given")
    return None


def scsi_wwid_of_iscsi_session(h_iqn=None, t_iqn=None, sid=None):
    """Usage
        scsi_wwid_of_iscsi_session();
        scsi_wwid_of_iscsi_session(sid=1);
        scsi_wwid_of_iscsi_session(h_iqn=h_iqn, t_iqn=t_iqn);
    # we should not support this method since the h_iqn for qla4xxx
    #    scsi_wwid_of_iscsi_session(t_iqn=t_iqn, h_iqn=h_iqn);
        scsi_wwid_of_iscsi_session(iface=iface,target_ip=target_ip,;
            t_iqn=t_iqn);
        scsi_wwid_of_iscsi_session(session_id=session_id);
    Purpose
        Query out all SCSI WWIDs for certain iscsi session.
    Parameter
        h_iqn                  # the IQN used by the host
        t_iqn                  # the IQN used by iscsi target
        sid                    # the iSCSI session ID
    Returns
        wwids
            or
        None
    """
    wwids = None
    if sid:
        sid = str(sid)
        session_info = query_iscsi_session(sid)
        if not session_info:
            return None
        if "disks" in list(session_info.keys()):
            if not wwids:
                wwids = []
            for scsi_name in list(session_info["disks"].keys()):
                wwid = session_info["disks"][scsi_name]["wwid"]
                if wwid and wwid not in wwids:
                    wwids.append(wwid)
            return wwids
        return None

    sessions = query_all_iscsi_sessions()
    if not sessions:
        return None

    if not h_iqn and not t_iqn:
        for sid in list(sessions.keys()):
            if "disks" in list(sessions[sid].keys()):
                if not wwids:
                    wwids = []
                for scsi_name in list(sessions[sid]["disks"].keys()):
                    wwid = libsan.host.scsi.wwid_of_disk(scsi_name)
                    if wwid and wwid not in wwids:
                        wwids.append(wwid)
        return wwids

    if h_iqn and t_iqn:
        for sid in list(sessions.keys()):
            if sessions[sid]["h_iqn"] == h_iqn and sessions[sid]["t_iqn"] == t_iqn:  # noqa: SIM102
                if "disks" in list(sessions[sid].keys()):
                    if not wwids:
                        wwids = []
                    for scsi_name in list(sessions[sid]["disks"].keys()):
                        wwid = libsan.host.scsi.wwid_of_disk(scsi_name)
                        if wwid and wwid not in wwids:
                            wwids.append(wwid)
        return wwids

    _print("FAIL: scsi_wwid_of_iscsi_session() - Unsupported parameters given")
    return None


def is_iscsi_boot():
    iscsi_wwids = scsi_wwid_of_iscsi_session()
    if not iscsi_wwids:
        return False
    boot_dev = libsan.host.linux.get_boot_device()
    if not boot_dev:
        _print("FAIL: is_iscsi_boot() - Could not determine boot device")
        return False

    boot_wwid = libsan.host.linux.get_device_wwid(boot_dev)
    if not boot_wwid:
        _print("WARN: is_iscsi_boot() - Could not determine boot WWID for %s" % boot_dev)
        return False

    if boot_wwid in iscsi_wwids:
        return True

    return False


# iSCSI node ###
def node_login(options=None, target=None, portal=None, udev_wait_time=15):
    """Login to an iSCSI portal, or all discovered portals
    The arguments are:
    \target:    iSCSI targets to be used, separated by space (optional)
    \toptions:   extra paramters. eg: "-T <target> -p <portal>"
    Returns:
    \tTrue:     If iSCSI node is logged in
    \tFalse:    If some problem happened
    """

    # Going to delete discovered target information
    _print("INFO: Performing iSCSI login")
    cmd = "iscsiadm -m node -l"
    if options:
        cmd += " %s" % options

    if target:
        for target_iqn in target.split():
            cmd += " -T %s" % target_iqn

    if portal:
        cmd += " -p %s" % portal

    retcode, output = run(cmd, return_output=True, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not login to iSCSI target")
        print(output)
        return False

    libsan.host.linux.wait_udev(udev_wait_time)
    return True


def node_logout(options=None, target=None, portal=None):
    """Logout from an iSCSI node
    The arguments are:
    \toptions:   extra paramters. eg: "-T <target> -p <portal>"
    Returns:
    \tTrue:     If iSCSI node is removed
    \tFalse:    If some problem happened
    """
    ses_dict = query_all_iscsi_sessions()
    if not ses_dict:
        # There is no session to logout just skip
        return True
    _print("INFO: Performing iSCSI logout")
    # Going to logout discovered target information
    cmd = "iscsiadm -m node -u"
    if options:
        cmd += " %s" % options

    if target:
        cmd += " -T %s" % target

    if portal:
        cmd += " -p %s" % portal

    retcode, output = run(cmd, return_output=True, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not logout from iSCSI target")
        print(output)
        return False
    return True


def node_delete(options=None):
    """Delete node information"""
    if not options:
        _print("FAIL: node_delete() - requires portal and/or target parameters")
        return False

    # Going to logout discovered target information
    cmd = "iscsiadm -m node -o delete"
    if options:
        cmd += " %s" % options

    retcode, _ = run(cmd, return_output=True, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not delete node iSCSI target")
        return False
    return True


# iSCSI iface ###
def iface_query_all_info(iface_name=None):
    """Return a dict with interface names as key with detailed information of
    interface
    """
    ifaces = [iface_name] if iface_name else get_iscsi_iface_names()

    if not ifaces:
        return None

    all_iface_dict = {}
    iface_info_regex = re.compile(r"iface\.(\S+) = (\S+)")

    for iface in ifaces:
        cmd = "iscsiadm -m iface -I %s" % iface
        retcode, output = run(cmd, return_output=True, verbose=False)
        if retcode != 0:
            _print("FAIL: Could not delete node iSCSI target")
            continue
        details = output.split("\n")
        for info in details:
            m = iface_info_regex.match(info)
            if not m:
                continue
            if iface not in list(all_iface_dict.keys()):
                all_iface_dict[iface] = {}
            value = m.group(2)
            if value == "<empty>":
                value = None
            all_iface_dict[iface][m.group(1)] = value

    if iface_name:
        if iface_name not in list(all_iface_dict.keys()):
            return None
        return all_iface_dict[iface_name]

    return all_iface_dict


def iface_update(iface, name, value):
    """Updates iSCSI interface parameter
    The arguments are:
    \tiface # Interface name (-I $)
    \tname  # Name of parameter (-n iface.$)
    \tvalue  # Value to set (-v $)
    Returns:
    \tTrue:     If value is set successfully
    \tFalse:    If some problem happened
    """
    if not iface or not name or not value:
        _print("FAIL: iface_update() - required parameters: iface, name, value")
        return False

    cmd = f"iscsiadm -m iface -I {iface} -o update -n iface.{name} -v {value}"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print(f"FAIL: Could not set {name} to {value} on iface {iface}")
        print(output)
        return False

    return True


def iface_set_iqn(iqn, iface="default"):
    """Set IQN in /etc/iscsi/initiatorname or for specific iface
    Return:
        True
        of
        False
    """
    if not iqn:
        _print("FAIL: iface_set_iqn() - requires iqn to be set")
        return False

    if iface == "default":
        try:
            with open("/etc/iscsi/initiatorname.iscsi", "w", encoding="UTF-8") as i:
                i.write("InitiatorName=" + str(iqn))
        except Exception as e:
            _print("FAIL: Could not set iqn in /etc/iscsi/initiatorname.iscsi " + e.__str__())
            return False
        libsan.host.linux.service_restart("iscsid")

        return True

    if not iface_update(iface, "initiatorname", iqn):
        return False

    return True


def iface_set_ip(iface, ip, mask=None, gw=None):
    """Set IP information for specific iface
    Return:
        True
        of
        False
    """
    if not iface or not ip:
        _print("FAIL: iface_set_ip() - requires iface and ip parameters")
        return False

    if not iface_update(iface, "ipaddress", ip):
        return False

    if mask and not iface_update(iface, "subnet_mask", mask):
        return False

    if gw and not iface_update(iface, "gateway", gw):
        return False

    return True


def get_iscsi_iface_names():
    """Return a list with the name of all iSCSI interfaces on the host"""
    cmd = 'iscsiadm -m iface | cut -d " " -f 1'
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: Could not read iSCSI interfaces")
        print(output)
        return None
    ifaces = output.split("\n")
    ifaces[:] = (value for value in ifaces if "iSCSI ERROR" not in value)  # bz1997710
    return ifaces


def set_iscsid_parameter(parameters):
    """Change parameter in iscsid.conf file and restarts iscsid service
    Use dictionary with parameter:value as argument
    """

    filename = "/etc/iscsi/iscsid.conf"
    if not libsan.host.linux.edit_config(filename, parameters, list_values=False):
        _print(f"FAIL: Unable to set iscsi parameters: {parameters}")
        return False

    if not libsan.host.linux.service_restart("iscsid"):
        _print("FAIL: Unable to restart iscsid service")
        return False

    return True


def set_chap(target_user, target_pass, initiator_user=None, initiator_pass=None):
    """Set CHAP authentication.
    Arguments:
    target_user, target_pass -- Username and password used for 1-way authentication.

    initiator_user, initiator pass -- optional, used for 2-way bidirectional authentication.
    """

    if not target_user or not target_pass:
        print("FAIL: set_chap() - requires username and password")
        return False

    parameters = {
        "node.session.auth.authmethod": "CHAP",
        "node.session.auth.username": target_user,
        "node.session.auth.password": target_pass,
        "discovery.sendtargets.auth.authmethod": "CHAP",  # NetApp array requires discovery authentication
        "discovery.sendtargets.auth.username": target_user,
        "discovery.sendtargets.auth.password": target_pass,
    }

    if initiator_user and initiator_pass:
        print("INFO: Setting mutual two-way CHAP authentication")
        parameters["node.session.auth.username_in"] = initiator_user
        parameters["node.session.auth.password_in"] = initiator_pass
        parameters["discovery.sendtargets.auth.username_in"] = initiator_user
        parameters["discovery.sendtargets.auth.password_in"] = initiator_pass

    if not set_iscsid_parameter(parameters):
        print("FAIL: Unable to set CHAP authentication")
        return False

    if not libsan.host.linux.service_restart("iscsid"):
        print("FAIL: Unable to restart iscsid service")
        return False

    print("INFO: CHAP authentication enabled")
    return True


def disable_chap():
    """Disable CHAP authentication in iscsid.conf and restarts the service"""

    # Removing all previously set auth parameters.
    parameters = [
        "node.session.auth.authmethod",
        "node.session.auth.username",
        "node.session.auth.password",
        "discovery.sendtargets.auth.authmethod",
        "discovery.sendtargets.auth.username",
        "discovery.sendtargets.auth.password",
        "node.session.auth.username_in",
        "node.session.auth.password_in",
        "discovery.sendtargets.auth.username_in",
        "discovery.sendtargets.auth.password_in",
    ]

    libsan.host.linux.remove_from_config("/etc/iscsi/iscsid.conf", parameters, warn=False)

    if not libsan.host.linux.service_restart("iscsid"):
        print("FAIL: Unable to restart iscsid service")
        return False

    return True


def multipath_timeo(seconds=None):
    """If multipath is used for iSCSI session, session replacement
    timeout time should be decreased from default 120 seconds
    https://access.redhat.com/solutions/1171203
    multipathd service should be running when calling this
    The arguments are:
    \tSeconds - default 10 or number of seconds
    Returns:
    \tTrue: Successfully modified iscsid config file.
    \tFalse: There was some problem.
    """

    param = "node.session.timeo.replacement_timeout"

    if not seconds:
        seconds = 10
    seconds = str(seconds)

    if libsan.host.mp.is_multipathd_running():
        print("INFO: multipathd is running")
    else:
        print("FAIL: multipathd is not running")
        return False

    if not set_iscsid_parameter({param: seconds}):
        return False

    return True


def create_iscsi_iface(iface_name, mac=None):
    """Create a new iSCSI interface, assign mac if specified"""
    if not iface_name:
        _print("FAIL: create_iscsi_iface() - requires iface name as parameter")
        return False

    if iface_name in get_iscsi_iface_names():
        _print("INFO: iSCSI interface '%s' already exists" % iface_name)
        return True

    cmd = "iscsiadm -m iface -o new -I %s" % iface_name
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: Could not create iSCSI interface")
        print(output)
        return False

    if mac is not None and not iface_update(iface_name, "hwaddress", mac):
        return False

    return True


def clone_iscsi_iface(new_iface_name, base_iface):
    print(f"Cloning iface: {base_iface} to {new_iface_name}")
    if not create_iscsi_iface(new_iface_name):
        return False

    iface_info = iface_query_all_info(base_iface)
    if iface_info is None:
        print("FAIL: Could not query all info about iface: %s" % base_iface)
        return False

    if iface_info["hwaddress"] is not None and not iface_update(new_iface_name, "hwaddress", iface_info["hwaddress"]):
        return False

    if iface_info["transport_name"] is not None:  # noqa: SIM102
        if not iface_update(new_iface_name, "transport_name", iface_info["transport_name"]):
            return False

    if iface_info["initiatorname"] is not None:  # noqa: SIM102
        if not iface_update(new_iface_name, "initiatorname", iface_info["initiatorname"]):
            return False

    if iface_info["ipaddress"] is not None and not iface_update(new_iface_name, "ipaddress", iface_info["ipaddress"]):
        return False

    print(f"successully cloned {base_iface}. new iface: {new_iface_name}")
    return True


def remove_iscsi_iface(iface_name):
    if iface_name not in get_iscsi_iface_names():
        _print("INFO: iSCSI interface '%s' does not exist" % iface_name)
        return False

    cmd = "iscsiadm -m iface -o delete -I %s" % iface_name
    retcode = run(cmd, verbose=False)
    if retcode != 0:
        _print("FAIL: Could not remove iSCSI interface")
        return False

    return True


def node_iface_info(iface_name):
    cmd = "iscsiadm -m node -I %s" % iface_name
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: Could not get iface info!")
        print(output)
        return False

    return True


# iSCSI disks ###


def get_all_iscsi_disks():
    sessions = query_all_iscsi_sessions()
    disks = []
    if not sessions:
        # there is no iSCSI session
        return None

    # search for disks in each session
    for sid in list(sessions.keys()):
        ses = sessions[sid]
        if ses["disks"]:
            # disk names are key values
            disks.extend(list(ses["disks"].keys()))

    return disks


def get_session_id_from_disk(disk_name):
    for sid in query_all_iscsi_sessions():
        session = query_iscsi_session(sid)
        if not session:
            _print("FAIL: Could not query iscsi session sid: '%s'." % sid)
        if disk_name in session["disks"]:
            return session["sid"]
    _print("FAIL: Could not find disk '%s' in iscsi sessions." % disk_name)
    return None



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""conf.py: Module to SAN config file."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import os
import re
from configparser import ConfigParser

from libsan import _print


def load_config(config_file):
    """Read a conf file
    The arguments are:
    \tconfig_file:         Full path to file where is the config
    Returns:
    \\tList:               A list with all config
    """
    if not os.path.isfile(config_file):
        _print("FAIL: Could not read %s" % config_file)
        return None

    # config = ConfigParser()
    # We want to preserve case
    # config.optionxform=str
    # config.read(config_file)

    # class_name_dict = {}
    # class_name_dict["switch"] = "libsan.switch"
    # class_name_dict["array"] = "libsan.array"
    # class_name_dict["pyswitch"] = "libsan.physwitch"

    config_dict = []

    # for section in config.sections():
    #    section_dict = {}
    #    section_dict["name"] = section
    #    section_dict["options"] = {}
    #    for option in config.options(section):
    #        sec_key = option
    #        #remove FC port prefix
    #        #sec_key = re.sub("^wwpn-", "", sec_key)
    #        value = config.get(section, option)
    #        #we need to trail comments out
    #        value = re.sub("#.*", "", value)
    #        value = value.strip()
    #        section_dict["options"][sec_key] = value
    #    config_dict.append(section_dict)

    # Not using ConfigParser as we would like to have repeated sections
    try:
        with open(config_file) as f:
            config_output = f.read()
    except Exception as e:
        print(e)
        _print("FAIL: load_config() - Could not read %s" % config_file)
        return None

    lines = config_output.split("\n")
    blank_line = re.compile(r"\s+")
    comment_regex = re.compile(r".*(#.*)$")
    section_regex = re.compile(r"\s*\[(.*)\]")
    option_regex = re.compile(r"\s*(\S+)\s*=\s*(.*)$")

    current_section = None
    sec_dict = None
    for line in lines:
        actual_line = line
        m = comment_regex.match(line)
        if m:
            # remove comment
            actual_line = actual_line.replace(m.group(1), "")
        if blank_line.match(actual_line):
            continue
        # Check if it is a section
        m = section_regex.match(actual_line)
        if m:
            # New section
            if current_section:
                # add previous section to config_dict
                config_dict.append(sec_dict)
            current_section = actual_line
            # creating new section data
            sec_dict = {}
            sec_dict["name"] = m.group(1)
            sec_dict["options"] = {}
            continue
        # Check if it is option of section
        m = option_regex.match(actual_line)
        if m:
            if not current_section:
                _print("FAIL: load_config() - option (%s) does not belong to any section" % line)
                return None
            # Store the value removing trailing spaces
            sec_dict["options"][m.group(1)] = m.group(2).strip()
    # add last section to config_dict
    if sec_dict:
        config_dict.append(sec_dict)
    return config_dict


def load_config_as_json(config_file):
    """Read a conf file
    The arguments are:
    \tconfig_file:         Full path to file where is the config
    Returns:
    \\tDict:               A dict with all config
    """
    if not os.path.isfile(config_file):
        _print("FAIL: Could not read %s" % config_file)
        return None

    config_dict = {}

    try:
        with open(config_file) as f:
            config_output = f.read()
    except Exception as e:
        print(e)
        _print("FAIL: load_config_as_json() - Could not read %s" % config_file)
        return None

    lines = config_output.split("\n")
    blank_line = re.compile(r"\s+")
    comment_regex = re.compile(r".*?(#+.*)$")
    section_regex = re.compile(r"\s*\[(.*)\]")
    option_regex = re.compile(r"\s*(\S+)\s*=\s*(.*)$")

    current_section = None
    for line in lines:
        actual_line = line
        m = comment_regex.match(line)
        if m:
            # remove comment
            actual_line = actual_line.replace(m.group(1), "")
        if blank_line.match(actual_line):
            continue
        # Check if it is a section
        m = section_regex.match(actual_line)
        if m:
            name = m.group(1)
            if name in config_dict:
                _print("FAIL: load_config_as_json() - section (%s) already exist" % name)
                return None
            config_dict[name] = {}
            current_section = name
            continue
        # Check if it is option of section
        m = option_regex.match(actual_line)
        if m:
            key = m.group(1)
            value = m.group(2).strip()
            if not current_section:
                _print("FAIL: load_config_as_json() - option (%s) does not belong to any section" % line)
                return None
            if key in config_dict[current_section]:
                _print(f"FAIL: load_config_as_json() - option ({key}) already exist in: {current_section}")
                return None
            # Store the value removing trailing spaces
            config_dict[current_section][key] = value

    return config_dict


def config_add_entry(config_key, config_param, param_value, config_file):
    """Add entry to a config file
    The arguments are:
    \tconfig_key:           Key register that we will add the entry
    \tconfig_param:         The config parameter that will be added
    \tconfig_value:         The value for the parameter
    \tconfig_file:          Full path to file where is the config
    Returns:
        True
        or
        False
    """

    if not os.path.isfile(config_file):
        _print("FAIL: config_add_entry() - Could not open %s" % config_file)
        return None

    config = ConfigParser()
    # We want to preserve case
    config.optionxform = str
    config.read(config_file)

    config.set(config_key, config_param, param_value)

    with open(config_file, "w", encoding="UTF-8") as configfile:
        config.write(configfile)

    return True


def config_remove_entry(config_key, config_param, config_file):
    """Remove entry from a config file
    The arguments are:
    \tconfig_key:           Key register that we will add the entry
    \tconfig_param:         The config parameter that will be added
    \tconfig_file:          Full path to file where is the config
    Returns:
        True
        or
        False
    """

    if not os.path.isfile(config_file):
        _print("FAIL: config_add_entry() - Could not open %s" % config_file)
        return None

    config = ConfigParser()
    # We want to preserve case
    config.optionxform = str
    config.read(config_file)

    config.remove_option(config_key, config_param)

    with open(config_file, "w", encoding="UTF-8") as configfile:
        config.write(configfile)

    return True


def get_alias(config_dict, value):
    """Given a value get its alias if exist.
    The arguments are:
    \tconfig_dict:        The config dict return by load_config
    \tvalue:              The value we want get the alias from
    Returns:
    \tString:             The alias
    \tNone:               There was a problem or alias does not exist
    """
    if "alias" not in list(config_dict.keys()):
        return None

    for alias in list(config_dict["alias"].keys()):
        if config_dict["alias"][alias] == value:
            return alias
    return None



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.


"""cmdline.py: Module to execute a command line."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import subprocess
import sys
import time
from shutil import which


def run(cmd, return_output=False, verbose=True, force_flush=False, timeout=None):
    """Run a command line specified as cmd.
    The arguments are:
    \tcmd (str):    Command to be executed
    \tverbose:      if we should show command output or not
    \tforce_flush:  if we want to show command output while command is being executed. eg. hba_test run
    \ttimeout (int):timeout for the process in seconds
    \treturn_output (Boolean): Set to True if want output result to be returned as tuple. Default is False
    Returns:
    \tint: Return code of the command executed
    \tstr: As tuple of return code if return_output is set to True"""
    # by default print command output
    if verbose:
        # Append time information to command
        date = 'date "+%Y-%m-%d %H:%M:%S"'
        p = subprocess.Popen(date, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = p.communicate()
        stdout = stdout.decode("ascii", "ignore")
        stdout = stdout.rstrip("\n")
        print(f"INFO: [{stdout}] Running: '{cmd}'...")

    stderr = b""
    if not force_flush:
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        try:
            stdout, stderr = p.communicate(timeout=timeout)
        except subprocess.TimeoutExpired:
            p.kill()
            stdout, stderr = p.communicate()
            print("WARN: Timeout reached")
            p.returncode = 124  # to stay consistent with bash Timeout return code
        sys.stdout.flush()
        sys.stderr.flush()
    else:
        start_time = time.time()
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
        stdout = b""
        while p.poll() is None:
            new_data = p.stdout.readline()
            stdout += new_data
            if verbose:
                sys.stdout.write(new_data.decode("ascii", "ignore"))
            sys.stdout.flush()
            if timeout and time.time() - start_time > timeout:
                print("WARN: Timeout reached")
                p.kill()
                p.returncode = 124  # to stay consistent with bash Timeout return code
                break

    retcode = p.returncode

    # print "stdout:(" + stdout + ")"
    # print "stderr:(" + stderr + ")"
    output = stdout.decode("ascii", "ignore") + stderr.decode("ascii", "ignore")

    # remove new line from last line
    output = output.rstrip()

    # by default print command output
    # if force_flush we already printed it
    if verbose and not force_flush:
        print(output)

    # print "stderr " + err
    # print "returncode: " + str(retcode)
    if not return_output:
        return retcode
    else:
        return retcode, output


def exists(cmd):
    if not which(str(cmd)):
        return False
    return True



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""ontap.py: Module to manager NetApp Ontap Array with API,
the system software should higher than 9.6.
"""

__author__ = "Guangwu Zhang"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import base64
import json
import socket
import time

import libsan.host.fc
import libsan.host.linux
import libsan.host.scsi
import libsan.misc.size
from libsan import _print
from netapp_ontap import HostConnection, NetAppRestError, config
from netapp_ontap.resources import (
    CLI,
    FcInterface,
    Igroup,
    IgroupInitiator,
    IpInterface,
    Lun,
    LunMap,
    Node,
    Software,
    Svm,
    Volume,
)


def _lun_serial2wwid(serial):
    """Usage
        _lun_serial2wwid(netapp_serial)
    Purpose
        Convert netapp serial to WWID.
        1. Convert serial to hex string
        2. add '360a98000' at the head
    Parameter
        $netapp_serial  # like '2FiCl+BOAef9'
    Returns
        wwid
    """
    if not serial:
        return None

    # toHex = lambda x:"".join([hex(ord(c))[2:].zfill(2) for c in x])
    wwid = serial

    # wwid = toHex(wwid)
    wwid = base64.b16encode(str.encode(wwid))
    wwid = "3600a0980" + wwid.decode()
    return wwid.lower()


class ontap:  # noqaN801
    def __init__(
        self,
        hostname,
        user,
        passwd,
        timeout=None,
        san_dev=None,
    ):
        self.hostname = hostname
        self.user = user
        self.passwd = passwd
        self.san_dev_name = san_dev
        self.luns = None
        self.igroup = None
        self.vol = None
        self.svm = None
        self.wwid = None
        self.san_conf_path = None
        self.sa_conf_dict = None
        config.CONNECTION = HostConnection(
            self.hostname, username=self.user, password=self.passwd, verify=False, poll_timeout=timeout
        )

    def set_san_conf_path(self, san_conf_path):
        self.san_conf_path = san_conf_path
        return True

    def set_sa_conf(self, sa_conf_dict):
        self.sa_conf_dict = sa_conf_dict
        return True

    def get_sa_conf(self):
        return self.sa_conf_dict

    def lun_info(self, lun_name):
        """Query detailed information of specific LUN"""
        if not lun_name:
            _print("FAIL: lun_info - requires lun_name")
            return None
        luns_dict = self.query_all_lun_info(recheck=True) if not self.luns or lun_name not in self.luns else self.luns

        if not luns_dict:
            _print("FAIL: lun_info() - Could not query all lun info")
            return None
        if lun_name in list(luns_dict.keys()):
            return luns_dict[lun_name]

        _print("INFO: lun_info(): Could not find LUN %s" % lun_name)
        return None

    def get_nodes(self, ip=None):
        """Node name: na-fas8080b
        Node name: na-fas8080a
        """
        nodes = {}

        try:
            for node in Node.get_collection():
                node.get()
                nodes[node.name] = {}
                nodes[node.name]["name"] = node.name
                nodes[node.name]["uuid"] = node.uuid
                nodes[node.name]["state"] = node.state
                nodes[node.name]["version"] = f"{node.version.generation}.{node.version.major}"
                nodes[node.name]["ip"] = node.management_interfaces[0].ip.address
                if socket.gethostbyname(ip) == node.management_interfaces[0].ip.address:
                    return nodes[node.name]

        except NetAppRestError as error:
            _print("FAIL: Error: %s" % error.http_err_response.http_response.text)
            _print("FAIL: Exception caught :" + str(error))
        return nodes

    def get_version(self):
        sw = Software()
        sw.get()
        return sw["version"]

    def get_svms(self):
        vms = {}
        try:
            for svm in Svm.get_collection(fields="uuid"):
                vms[svm.name] = svm.uuid
        except NetAppRestError as error:
            _print("FAIL: Error:- %s" % error.http_err_response.http_response.text)
            _print("FAIL: Exception caught :" + str(error))
        return vms

    def get_vols(self, vm_name):
        vols = {}
        try:
            for volume in Volume.get_collection(**{"svm.name": vm_name}, fields="uuid"):
                vols[volume.name] = volume.uuid
        except NetAppRestError as error:
            _print("FAIL: Error:- %s" % error.http_err_response.http_response.text)
            _print("FAIL: Exception caught :" + str(error))
        return vols

    def get_igroups(self, vm_name, igroupname=None, h_wwpn=None):
        igroups = {}
        try:
            for igroup in Igroup.get_collection(**{"svm.name": vm_name}, fields="uuid"):
                igroup.get()
                igroups[igroup.name] = {}
                igroups[igroup.name]["name"] = igroup.name
                igroups[igroup.name]["uuid"] = igroup.uuid
                igroups[igroup.name]["os_type"] = igroup.os_type
                igroups[igroup.name]["svm"] = {igroup.svm.name: igroup.svm.uuid}
                igroups[igroup.name]["type"] = igroup.protocol
                igroups[igroup.name]["delete_on_unmap"] = igroup.delete_on_unmap
                igroups[igroup.name]["members"] = [init.name for init in igroup.initiators]

                if vm_name == igroup.svm.name:
                    if isinstance(h_wwpn, list) and all(wwpn in igroups[igroup.name]["members"] for wwpn in h_wwpn):
                        return igroups[igroup.name]
                    if igroupname == igroup.name:
                        return igroups[igroup.name]

        except NetAppRestError as error:
            _print("FAIL: Error:- %s" % error.http_err_response.http_response.text)
            _print("FAIL: Exception caught :" + str(error))
        return igroups

    def get_h_wwpn(
        self,
    ):
        return libsan.host.fc.h_wwpn_of_host()

    def get_initiators(self, vm_name, igroup_name):
        init_name = []
        igroup_uuid = ""
        igroup = self.get_igroups(vm_name, igroup_name)
        if "uuid" in igroup:
            igroup_uuid = igroup["uuid"]
        else:
            _print(f"FAIL: Could not find the igroup uuid with {igroup_name} in svm {vm_name}")
            return init_name

        try:
            for ini in IgroupInitiator.get_collection(igroup_uuid):
                init_name.append(ini.name)

        except NetAppRestError as error:
            _print("FAIL: Error:- %s " % error.http_err_response.http_response.text)
            _print("FAIL: Exception caught :" + str(error))
        return init_name

    @staticmethod
    def capability():
        """Indicates supported operation on array"""
        cap_dict = {}
        cap_dict["lun_info"] = True
        cap_dict["lun_query"] = True
        cap_dict["lun_create"] = True
        cap_dict["lun_map"] = True
        cap_dict["lun_unmap"] = True
        cap_dict["lun_remove"] = True
        cap_dict["lun_grow"] = False
        cap_dict["lun_shrink"] = False
        cap_dict["lun_trepass"] = False
        cap_dict["lun_thinp"] = False
        cap_dict["sa_ctrler_reboot"] = True
        return cap_dict

    def get_lun_from_wwid(self, wwid):
        all_luns = self.query_all_lun_info()
        for lun_name in list(all_luns.keys()):
            if all_luns[lun_name]["wwid"] == wwid and all_luns[lun_name]["mapped"]:
                self.svm = all_luns[lun_name]["svm"]
                self.vol = all_luns[lun_name]["vol"]
                self.igroup = all_luns[lun_name]["map_infos"][0]["igroup_name"]
                return all_luns[lun_name]
        return None

    def query_all_luns(self, recheck=False):
        return sorted(self.query_all_lun_info(recheck=recheck).keys())

    def query_all_lun_info(self, recheck=False):
        """Lun Name:- /vol/fc_vol_5/dynamic_lun; Lun UUID:- 6668c02e-a0c1-490f-a932-e8e998d72731
        Lun Name:- /vol/tmp_luns_vol/lun0; Lun UUID:- 99be6020-b952-467c-b634-ffe4c994bf13
        """
        luns = {}
        if self.luns and not recheck:
            return self.luns

        map_luns = self.get_map_luns()
        try:
            for lun in Lun.get_collection():
                lun.get()
                luns[lun.name] = {}
                luns[lun.name]["uuid"] = lun["uuid"]
                luns[lun.name]["name"] = lun["name"]
                luns[lun.name]["os_type"] = lun["os_type"]
                luns[lun.name]["create_time"] = lun["create_time"]
                luns[lun.name]["sname"] = lun["location"]["logical_unit"]
                luns[lun.name]["vol"] = {lun["location"]["volume"]["name"]: lun["location"]["volume"]["uuid"]}
                luns[lun.name]["size"] = lun["space"]["size"]
                luns[lun.name]["size_human"] = libsan.misc.size.size_bytes_2_size_human(lun["space"]["size"])
                luns[lun.name]["size_netapp"] = lun["space"]["size"]
                luns[lun.name]["serial"] = lun["serial_number"]
                luns[lun.name]["wwid"] = _lun_serial2wwid(lun.serial_number)
                luns[lun.name]["status"] = lun["status"]["state"]
                luns[lun.name]["svm"] = {lun["svm"]["name"]: lun["svm"]["uuid"]}
                luns[lun.name]["enabled"] = lun["enabled"]
                luns[lun.name]["map_infos"] = []
                if lun.name in map_luns and lun.uuid == map_luns[lun.name]["uuid"]:
                    luns[lun.name]["mapped"] = True
                    lun_id = map_luns[lun.name]["lun_id"]
                    igroup_name = map_luns[lun.name]["igroup"]

                    for init in self.get_initiators(lun["svm"]["name"], igroup_name):
                        map_info = {
                            "igroup_name": igroup_name,
                            "t_lun_id": lun_id,
                            "h_wwpn": init,
                        }
                        luns[lun.name]["map_infos"].append(map_info)

        except NetAppRestError as error:
            _print("FAIL: Error:- %s" % error.http_err_response.http_response.text)
            _print("FAIL: Exception caught :" + str(error))
        self.luns = luns
        return luns

    def get_map_luns(self, lun_name=None):
        luns = {}
        try:
            for mlun in LunMap.get_collection():
                mlun.get()
                luns[mlun.lun.name] = {}
                luns[mlun.lun.name]["name"] = mlun.lun.name
                luns[mlun.lun.name]["uuid"] = mlun.lun.uuid
                luns[mlun.lun.name]["lun_id"] = mlun.logical_unit_number
                luns[mlun.lun.name]["node"] = mlun.lun.node.name
                luns[mlun.lun.name]["svm"] = mlun.svm.name
                luns[mlun.lun.name]["igroup"] = mlun.igroup.name

                if lun_name == mlun.lun.name:
                    return luns[mlun.lun.name]

        except NetAppRestError as error:
            _print("FAIL: Error:- %s" % error.http_err_response.http_response.text)
            _print("FAIL: Exception caught :" + str(error))
        return luns

    def get_interfaces(self):
        all_interface = {}
        try:
            for interface in IpInterface.get_collection():
                interface.get()
                all_interface[interface.name] = {}
                all_interface[interface.name]["name"] = interface.name
                all_interface[interface.name]["uuid"] = interface.uuid

        except NetAppRestError as error:
            _print("FAIL: Error:- %s" % error.http_err_response.http_response.text)
            _print("FAIL: Exception caught :" + str(error))
        return all_interface

    def get_fc_interface(
        self,
        svm=None,
        interface_name=None,
    ):
        target_lif_fc_ports = {}
        try:
            for port in FcInterface.get_collection():
                port.get()
                target_lif_fc_ports[port.name] = {}
                target_lif_fc_ports[port.name]["name"] = port.name
                target_lif_fc_ports[port.name]["uuid"] = port.uuid
                target_lif_fc_ports[port.name]["wwpn"] = port.wwpn
                target_lif_fc_ports[port.name]["wwnn"] = port.wwnn
                target_lif_fc_ports[port.name]["state"] = port.state
                target_lif_fc_ports[port.name]["location"] = {
                    "node": port.location.port.node.name,
                    "port": port.location.port.name,
                }
                target_lif_fc_ports[port.name]["data_protocol"] = port.data_protocol
                target_lif_fc_ports[port.name]["svm"] = {port.svm.name: port.svm.uuid}
                target_lif_fc_ports[port.name]["enabled"] = port.enabled

                if interface_name and interface_name == port.name:
                    return target_lif_fc_ports[port.name]

        except NetAppRestError as error:
            _print("FAIL: Error:- %s" % error.http_err_response.http_response.text)
            _print("FAIL: Exception caught :" + str(error))

        if svm:
            for port in list(target_lif_fc_ports.keys()):
                if svm not in target_lif_fc_ports[port]["svm"]:
                    target_lif_fc_ports.pop(port)
        return target_lif_fc_ports

    def mod_lif_fc_port(self, lif_name, enable=True):
        try:
            lif_fc_port = FcInterface.find(name=lif_name)
            lif_fc_port.enabled = enable
            if lif_fc_port.patch(poll=True):
                _print(f"INFO: update the lif fc port {lif_name} to {enable}")
        except NetAppRestError as error:
            _print("FAIL: Error:- %s" % error.http_err_response.http_response.text)
            _print("FAIL: Exception caught :" + str(error))
            return False
        return True

    def lun_create(self, name, size, os_type="linux", wwid=None):
        if wwid or self.wwid:
            self.get_lun_from_wwid(wwid)
        svm = self.svm if self.svm else self.get_sa_conf()["svm"]

        vol = self.vol if self.vol else self.get_sa_conf()["vol"]

        lun_full_path = "/vol/" + vol + "/" + name if "/vol/" + vol not in name else name
        sname = lun_full_path.split("/")[-1]
        _print(f"INFO: will crate new lun with vol {vol}  verver {svm}")
        l_size = libsan.misc.size.size_human_2_size_bytes(size)
        if not l_size:
            _print("FAIL: Could not convert %s to bytes" % size)
            return False
        try:
            vserver = Svm.find(name=svm)
            if not vserver:
                _print("FAIL: Could not find the vserver %s in the array" % svm)
                return False
            volume = Volume.find(name=vol, **{"svm.name": svm})
            if not volume:
                _print("FAIL: Could not find the volume %s in the array" % vol)
                return False
        except NetAppRestError as error:
            _print("FAIL: Exception caught :" + str(error))
            return False

        try:
            lun = Lun.find(name=lun_full_path)
            if lun:
                time.sleep(2)
                _print("INFO: The lun name %s have used" % lun_full_path)
                new_name = libsan.host.linux.time_stamp()
                lun_full_path += new_name
                sname += new_name
                _print("INFO: Set new lun name %s" % lun_full_path)
        except NetAppRestError as error:
            _print("FAIL: Exception caught :" + str(error))
            return False

        new_lun = {
            "comment": sname,
            "location": {"logical_unit": sname, "volume": {"name": vol}},
            "name": lun_full_path,
            "os_type": os_type,
            "space": {"guarantee": {"requested": bool("")}, "size": l_size},
            "svm": {"name": svm},
        }
        print(new_lun)
        lun_object = Lun.from_dict(new_lun)
        try:
            if lun_object.post(poll=True):
                _print("INFO :LUN created  %s created Successfully" % lun_object.name)
        except NetAppRestError as error:
            _print("FAIL: Exception caught :" + str(error))
            return False
        return lun_full_path

    def lun_remove(self, lun_name):
        self.lun_unmap(lun_name)
        try:
            lun = Lun.find(name=lun_name)
            if lun.delete(poll=True):
                _print("INFO :LUN deleted successfully.")
        except NetAppRestError as error:
            _print("FAIL: Exception caught :" + str(error))
            return False
        return True

    def lun_map(
        self,
        lun_full_name,
    ):
        h_wwpns = self.get_h_wwpn()

        svm = self.svm if self.svm else self.get_sa_conf()["svm"]

        if self.igroup:
            igroup_name = self.igroup
        else:
            igroup = self.get_igroups(svm, h_wwpn=h_wwpns)
            if igroup:
                igroup_name = igroup["name"]
            else:
                _print("FAIL: Could not find the igroup for the wwpn %s" % h_wwpns)
                return False

        if not igroup_name:
            _print("FAIL: Don't create igroup for th wwpn %s" % h_wwpns)
            return False

        map_info = {
            "svm": {"name": svm},
            "igroup": {"name": igroup_name},
            "lun": {"name": lun_full_name},
        }
        print("map info %s" % map_info)
        lun_object = LunMap.from_dict(map_info)
        try:
            if lun_object.post(poll=True):
                time.sleep(3)
                _print("INFO :Send lun map request to the array")
        except NetAppRestError as error:
            _print("FAIL: Exception caught :" + str(error))
            return False
        # after send mapped request, but the system don't recognize at once
        lun_info = self.lun_info(lun_full_name)
        if lun_info:
            n = 10
            while n > 0:
                for wwpn in self.get_initiators(svm, igroup_name):
                    host = libsan.host.fc.fc_host_id_of_wwpn(wwpn)
                    libsan.host.linux.run("echo 1 >/sys/class/fc_host/host%s/issue_lip" % host, False, False)
                    time.sleep(5)
                    libsan.host.scsi.rescan_host(host)
                    libsan.host.linux.wait_udev()
                if libsan.host.scsi.scsi_ids_of_wwid(lun_info["wwid"]):
                    break
                n -= 1

        else:
            _print("FAIL: Could not find the lun %s in the array" % lun_full_name)
            return False
        _print("INFO :LUN map Successfully")
        return True

    def lun_unmap(
        self,
        lun_name,
    ):
        map_luns = self.get_map_luns(lun_name)
        if "igroup" in map_luns:
            igroup_name = map_luns["igroup"]
        else:
            all_luns = self.query_all_lun_info(recheck=True)
            igroup_name = all_luns[lun_name]["map_infos"][0]["igroup_name"]
        try:
            lunmap = LunMap.find(lun=lun_name, igroup=igroup_name)
            if lunmap.delete(poll=True):
                _print("INFO: LUN unmap successfully.")
        except NetAppRestError as error:
            _print("FAIL: Exception caught :" + str(error))
            return False
        return True

    def sa_ctrler_t_wwpns(
        self,
        ctrler_ip,
    ):
        svm_ctrler_wwpn = {}
        node_name = self.get_nodes(ctrler_ip)
        svm = self.get_sa_conf()["svm"]
        all_fc_interface = self.get_fc_interface(svm)
        _print(f"DEBUG: vsm {svm} FC interface info {all_fc_interface}")
        _print(f"DEBUG: the IP {ctrler_ip} is from node name {node_name}")
        for lif in all_fc_interface:
            svm_ctrler_wwpn[lif] = []
            if all_fc_interface[lif]["location"]["node"] == node_name["name"]:
                svm_ctrler_wwpn[lif].append(all_fc_interface[lif]["wwpn"])
            else:
                svm_ctrler_wwpn.pop(lif)
        return svm_ctrler_wwpn

    def system_node_power_on(self, node_name):
        """system node power on module"""
        response = CLI().execute("system node power on", body={"node": node_name}, privilege_level="diagnostic")
        time.sleep(3)
        response = CLI().execute("system node power show", status="on")
        _print("DEBUG: %s" % json.dumps(response.http_response.json(), indent=4))

    def system_node_power_off(self, node_name):
        response = CLI().execute("system node power off", body={"node": node_name}, privilege_level="diagnostic")
        time.sleep(3)
        response = CLI().execute("system node power show", status="off")
        _print("DEBUG: %s" % json.dumps(response.http_response.json(), indent=4))

    def sa_ctrler_reboot(self, ctrler_ip):
        # na-fas8080a'
        # node_name = self.get_nodes(ctrler_ip)['name']
        # print(node_name)
        # self.system_node_power_off(node_name)
        # time.sleep(3)
        # self.system_node_power_on(node_name)
        # time.sleep(3)
        # return True

        for lif in self.sa_ctrler_t_wwpns(ctrler_ip):
            self.mod_lif_fc_port(lif, enable=False)
            time.sleep(2)
            self.mod_lif_fc_port(lif, enable=False)
        time.sleep(15)
        for lif in self.sa_ctrler_t_wwpns(ctrler_ip):
            self.mod_lif_fc_port(lif, enable=True)
            time.sleep(2)
            self.mod_lif_fc_port(lif, enable=True)

    def sa_ctrler_check(self, ctrler_ip):
        # na-fas8080b
        # na-fas8080a
        # response = CLI().execute(
        #     "system node power show", status="on")
        # print(response.http_response.json())
        # if int(json.dumps(response.http_response.json()['num_records'])) != 2 :
        #     return False
        # return True

        for lif in self.sa_ctrler_t_wwpns(ctrler_ip):
            if self.get_fc_interface(interface_name=lif)["enabled"]:
                return "online"
            return "offline"
